#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

import curses
import sys
import io
import logging

logger = logging.getLogger(__name__)

help_msg = "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\n" * 30 + "EOF"

text_example = (
    """\
What IS lOREM Ipsum? lOREM Ipsum IS simply dummy text of the printing and
typesetting industry. lOREM Ipsum has been the industry's standard dummy text
ever since the 1500s, when an unknown printer took a galley of type and
scrambled it to make a type specimen book. It has survived not only five
centuries, but also the leap into electronic typesetting, remaining essentially
unchanged. It was popularISed in the 1960s with the release of Letraset sheets
containing lOREM Ipsum passages, and more recently with desktop publIShing
software like Aldus PageMaker including versions of lOREM Ipsum.

Why do we use it? It IS a long establIShed fact that a reader will be
dIStracted by the readable content of a page when looking at its layout. The
point of using lOREM Ipsum IS that it has a more-or-less normal dIStribution of
letters, as opposed to using 'Content here, content here', making it look like
readable EnglISh. Many desktop publIShing packages and web page editors now use
lOREM Ipsum as their default model text, and a search for 'lorem ipsum' will
uncover many web sites still in their infancy. Various versions have evolved
over the years, sometimes by accident, sometimes on purpose (injected humour
and the like).

Where does it come from? Contrary to popular belief, lOREM Ipsum IS not simply
random text. It has roots in a piece of classical Latin literature from 45 BC,
making it over 2000 years old. Richard McClintock, a Latin professor at
Hampden-Sydney College in Virginia, looked up one of the more obscure Latin
words, consectetur, from a lOREM Ipsum passage, and going through the cites of
the word in classical literature, dIScovered the undoubtable source. lOREM
Ipsum comes from sections 1.10.32 and 1.10.33 of "de Finibus Bonorum et
Malorum" (The Extremes of Good and Evil) by Cicero, written in 45 BC. ThIS book
IS a treatISe on the theory of ethics, very popular during the RenaISsance. The
first line of lOREM Ipsum, "lOREM ipsum dolor sit amet..", comes from a line in
section 1.10.32.


The standard chunk of lOREM Ipsum used since the 1500s IS reproduced below for
those interested. Sections 1.10.32 and 1.10.33 from "de Finibus Bonorum et
Malorum" by Cicero are also reproduced in their exact original form,
accompanied by EnglISh versions from the 1914 translation by H. Rackham.

Where can I get some? There are many variations of passages of lOREM Ipsum
available, but the majority have suffered alteration in some form, by injected
humour, or randomISed words which don't look even slightly believable. If you
are going to use a passage of lOREM Ipsum, you need to be sure there ISn't
anything embarrassing hidden in the middle of text. All the lOREM Ipsum
generators on the Internet tend to repeat predefined chunks as necessary,
making thIS the first true generator on the Internet. It uses a dictionary of
over 200 Latin words, combined with a handful of model sentence structures, to
generate lOREM Ipsum which looks reasonable. The generated lOREM Ipsum IS
therefore always free from repetition, injected humour, or non-characterIStic
words etc.
        8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES
        8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES
\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB
\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB
\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB
\t\t2TABS\t\t2TABS\t\t2TABS
\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS
\t\t2TABS\t\t2TABS\t\t2TABS
\t\t2TABS\t\t2TABS\t\t2TAB
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\r
\aBELL
BELLS\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a
X\bBS
\fFF
\nNL
\rCR
\tHT
\vVT
\004EOT (EOF)
\033ESC
EOF_LINE
"""
    + """\
        8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES 8SPACES
\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB\t1TAB
\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TABS\t\t2TAB
\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL\aBELL
BELLS_FOLLOW
\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a\a
"""
)
# just avoid null


class TUI:
    """Curses class to handle text"""

    def __init__(self, text):
        # keep NL
        # self.virt_lines = text.splitlines(keepends=True)
        # remove NL
        self.virt_lines = text.split("\n")
        return

    ####################################################################
    # Externally used main method and effective main method
    ####################################################################
    # Design concept:
    #     * Use full terminal screen
    #     * Access them directly (no window nor pad)
    #     * String operation by pointers
    #
    # See https://docs.python.org/3/library/curses.html
    #     https://docs.python.org/3/howto/curses.html
    #
    #     curses terminal screen absolute coordinate as (y,x):
    #     +----------------------------------------------------------+
    #     | (0,0) top-left corner                    (term_col_max-1)|
    #     |  +--> x-axis                                             |
    #     |  |                                                       |
    #     |  |                 TEXT DATA DISPLAY                     |
    #     |  V y-axis                                                |
    #     |                                                          |
    #     |                                                          |
    #     | (term_row_max-2,0)        (term_row_max-2,term_col_max-1)|
    #     +----------------------------------------------------------+
    #     | (term_row_max-1,0) STATUS (term_row_max-1,term_col_max-1)|
    #     +----------------------------------------------------------+
    #
    # Here:
    #     * term_row_max, term_col_max = stdscr.getmaxyx()
    #
    # Internal data:
    #   * TUI
    #     * virt_to_row[virt_row] = (chunk_index, chunk_subindex)
    # Terminal size: 80 col x 24 row required
    #
    ####################################################################

    def main(self):  # for curses TUI (wrapper)
        """CURSES TUI program"""
        logger.debug("tui-main start")
        curses.wrapper(self.tui_loop)
        logger.debug("tui-main end")
        return

    def tui_loop(self, stdscr):  # for curses TUI (core)
        """wrapped CURSES TUI loop"""
        self.stdscr = stdscr
        self.init_curses()
        # initial display parameters
        # flag_update_text_win_corner = True
        corner_virt_row = 0
        corner_virt_col = 0
        while True:
            # enforce sanity
            corner_virt_col = max(0, corner_virt_col)
            corner_virt_row = max(0, corner_virt_row)
            self.display_imediff(corner_virt_row, corner_virt_col)
            c = self.stdscr.getch()  # return integer
            ch = chr(c)
            if ch == "w" or ch == "x" or c == curses.KEY_EXIT or c == curses.KEY_SAVE:
                ret = self.confirm_and_write_to_file()
                if ret == 0:
                    # success
                    break
                elif ret == 1:
                    # error
                    sys.exit(1)
                else:  # ret = -1
                    # skip and continue
                    pass
            elif ch == "q":
                logger.debug("quit without write to a file: imediff.out")
                break
            # elif ch == "h" or c == curses.KEY_HELP:
            # Show help screen
            # self.display_popup_win(self.get_helptext())
            elif ch == "H":
                # Show tutorial screen
                self.display_popup_win(help_msg)
            # Moves in document
            elif c == curses.KEY_HOME or ch == "0":
                corner_virt_row = 0
                corner_virt_col = 0
            elif c == curses.KEY_SF or c == curses.KEY_DOWN or ch == "j":
                corner_virt_row += 1
            elif c == curses.KEY_SR or c == curses.KEY_UP or ch == "k":
                corner_virt_row -= 1
            elif c == curses.KEY_NPAGE or ch == ";":
                corner_virt_row += 20
            elif c == curses.KEY_PPAGE or ch == "'":
                corner_virt_row -= 20
            elif c == curses.KEY_RIGHT or ch == "l":
                corner_virt_col += 8
            elif c == curses.KEY_LEFT or ch == "h":
                corner_virt_col -= 8
            # Terminal resize signal
            else:
                pass
            logger.debug("gui-command-loop")
        return

    ####################################################################
    def init_curses(self):
        # curses.start_color() # wrapper takes care
        logger.debug(
            "COLORS={}, COLOR_PAIRS={}".format(curses.COLORS, curses.COLOR_PAIRS)
        )
        curses.curs_set(0)  # cursor off
        # create text_win and status_win
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        logger.debug(
            "stdscr_row_max={}, stdscr_col_max={}".format(
                stdscr_row_max, stdscr_col_max
            )
        )
        if stdscr_row_max < 20 or stdscr_col_max < 60:
            logger.error(
                "E: terminal size too small: row={}:col={}".format(
                    stdscr_row_max, stdscr_col_max
                )
            )
            sys.exit(2)
        self.stdscr.clear()
        self.stdscr.refresh()
        # set up color
        color_a = "GREEN"  # a (OLDER/YOURS)
        color_b = "YELLOW"  # b (NEWER/COMMON)
        color_c = "CYAN"  # c (-----/THEIRS)
        color_m = "RED"  # marker d,f,null
        color_e = "MAGENTA"  # e (EDITOR)
        color_0 = "WHITE"  # tag=Ee attrib=DIM/NORMAL
        cc = dict()
        cc["GREEN"] = curses.COLOR_GREEN
        cc["YELLOW"] = curses.COLOR_YELLOW
        cc["CYAN"] = curses.COLOR_CYAN
        cc["RED"] = curses.COLOR_RED
        cc["MAGENTA"] = curses.COLOR_MAGENTA
        cc["BLUE"] = curses.COLOR_BLUE
        cc["WHITE"] = curses.COLOR_WHITE
        cc["BLACK"] = curses.COLOR_BLACK
        # set color pair_number as (pair_number, fg, bg)
        curses.init_pair(1, cc[color_a], cc["BLACK"])
        curses.init_pair(2, cc[color_b], cc["BLACK"])
        curses.init_pair(3, cc[color_c], cc["BLACK"])
        curses.init_pair(4, cc[color_m], cc["BLACK"])
        curses.init_pair(5, cc[color_e], cc["BLACK"])
        curses.init_pair(6, cc[color_0], cc["BLACK"])

    ####################################################################
    def display_text_to_a_row(
        self, win, row, col, line, i_b, i_e, attrib, clrtoeol=False
    ):
        #
        # win:   parent window
        # row:   row to put starting cursor
        # col:   col to put starting cursor
        # line:  input line
        # i_b:   input line begin point index (=)
        # i_e:   input line end point index (<)
        # attrib: curses attribute
        #
        # incoming must guarantee
        # * i_b = max(0, i_b)
        # * i_e = max(0, i_e)
        win_row_max, win_col_max = win.getmaxyx()
        # set starting cursor position robustly
        row0 = 0
        col0 = 0
        try:
            row0, col0 = win.getyx()
        except curses.error as _:
            logger.warning("win.getyx() cause error")
        if row is None:
            row = row0
        if col is None:
            col = col0
        # make sure to be 0 or plus
        row = max(0, row)
        col = max(0, col)
        # use always positive index
        if i_b is None:
            i_b = len(line)
        elif i_b < 0:
            i_b = max(0, len(line) - i_b)
        elif i_b >= len(line):
            logger.warning("i_b={} >= len(line)={}".format(i_b, len(line)))
            i_b = len(line)
        if i_e is None:
            i_e = len(line)
        elif i_e < 0:
            i_e = max(0, len(line) - i_e)
        elif i_e > len(line):
            logger.warning("i_e={} > len(line)={}".format(i_e, len(line)))
            i_e = len(line)
        if i_b > i_e:
            logger.warning("i_b={} > i_e={}".format(i_b, i_e))
            i_b = i_e
        # check reasonable cursor position
        if row < win_row_max and col < win_col_max:
            # single row window
            win1row = win.derwin(1, win_col_max, row, 0)
            win1row.move(0, col)
            # cursor starting point is not unreasonable
            try:
                win1row.addstr(line[i_b:i_e], attrib)
            except curses.error as _:
                pass
            if clrtoeol:
                win1row.clrtoeol()
            win1row.refresh()
        else:
            # not in writable cursor position
            pass

    def display_imediff(self, corner_virt_row, corner_virt_col):
        # text data 0 ................... < stdscr_row_max - 1
        # stat data stdscr_row_max -1 ... < stdscr_row_max
        row_max, col_max = self.stdscr.getmaxyx()
        for row_index in range(row_max - 1):
            virt_row_index = row_index + corner_virt_row
            if virt_row_index < len(self.virt_lines):
                line = self.virt_lines[virt_row_index].rstrip()
                self.display_text_to_a_row(
                    self.stdscr,
                    row_index,
                    0,
                    line,
                    corner_virt_col,
                    corner_virt_col + col_max,
                    self.get_attrib(virt_row_index),
                    clrtoeol=True,
                )
            else:
                self.display_text_to_a_row(
                    self.stdscr,
                    row_index,
                    0,
                    "...[EOF]...",
                    0,
                    col_max,
                    curses.color_pair(4) | curses.A_DIM,
                    clrtoeol=True,
                )
        status_line = "corner=(row{}:col{})".format(corner_virt_row, corner_virt_col)
        self.display_text_to_a_row(
            self.stdscr,
            row_max - 1,
            0,
            status_line,
            0,
            None,
            curses.color_pair(4) | curses.A_NORMAL,
            clrtoeol=True,
        )

    def get_attrib(self, virt_row_index):
        if virt_row_index < 6:
            attrib = curses.color_pair(0) | curses.A_BOLD
        else:
            attrib = curses.color_pair(virt_row_index % 6 + 1) | curses.A_NORMAL
        return attrib

    def display_popup_win(
        self, msg, msg_row_max=0, msg_col_max=0, margin_row=1, margin_col=2
    ):
        # A pop-up centered window with specified window size and border box
        # drawing.  The text are drawn in the box and has 1 space before and after
        # for readability.
        #
        # get msg display size for virt_msg
        virt_lines = msg.split("\n")  # no trailing \n
        virt_row_max = len(virt_lines)
        virt_col_max = 12  # enough for EOF amrker
        for line in virt_lines:
            virt_col_max = max(virt_col_max, len(line))
        # set up centered box_win
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        # usable msg display span
        span_col = stdscr_col_max - 2 * margin_col
        span_row = stdscr_row_max - 2 * margin_row
        logger.debug(
            "stdscr={}:{} virt={}:{} span={}:{}".format(
                stdscr_row_max,
                stdscr_col_max,
                virt_row_max,
                virt_col_max,
                span_row,
                span_col,
            )
        )
        if msg_row_max == 0 and (virt_row_max + 2 * margin_row) <= stdscr_row_max:
            msg_row_max = virt_row_max
        elif msg_row_max == 0 or msg_row_max >= span_row:
            # force reasonable msg_row_max
            msg_row_max = stdscr_row_max - 2 * margin_row
        if msg_col_max == 0 and (virt_row_max + 2 * margin_row) <= stdscr_row_max:
            msg_col_max = virt_col_max
        elif msg_col_max == 0 or msg_col_max >= span_col:
            # force reasonable msg_col_max
            msg_col_max = stdscr_col_max - 2 * margin_col
        box_margin_row = (stdscr_row_max - (msg_row_max + 2 * margin_row)) // 2
        box_margin_col = (stdscr_col_max - (msg_col_max + 2 * margin_col)) // 2
        self.box_win = self.stdscr.derwin(
            msg_row_max + 2 * margin_row,
            msg_col_max + 2 * margin_col,
            box_margin_row,
            box_margin_col,
        )
        self.box_win.clear()
        self.box_win.border()
        self.box_win.leaveok(True)
        self.box_win.refresh()
        # set up centered msg_win
        self.msg_win = self.box_win.derwin(
            msg_row_max, msg_col_max, margin_row, margin_col
        )
        corner_popup_row = 0
        corner_popup_col = 0
        while True:
            row_max, col_max = self.msg_win.getmaxyx()
            corner_popup_row = max(0, corner_popup_row)
            corner_popup_col = max(0, corner_popup_col)
            self.msg_win.move(0, 0)
            row_max, col_max = self.msg_win.getmaxyx()
            for row_index in range(row_max):
                virt_row_index = row_index + corner_popup_row
                if virt_row_index < len(virt_lines):
                    line = virt_lines[virt_row_index].rstrip()
                    self.display_text_to_a_row(
                        self.msg_win,
                        row_index,
                        0,
                        line,
                        corner_popup_col,
                        corner_popup_col + col_max,
                        curses.color_pair(virt_row_index % 6 + 1) | curses.A_NORMAL,
                        clrtoeol=True,
                    )
                else:
                    self.display_text_to_a_row(
                        self.msg_win,
                        row_index,
                        0,
                        "...[EOF]...",
                        0,
                        col_max,
                        curses.color_pair(4) | curses.A_DIM,
                        clrtoeol=True,
                    )
            self.msg_win.refresh()
            c = self.stdscr.getch()  # c : integer (stdscr! here)
            ch = chr(c)
            if c == curses.KEY_HOME or ch == "0":
                corner_popup_row = 0
                corner_popup_col = 0
            elif c == curses.KEY_SF or c == curses.KEY_DOWN or ch == "j":
                corner_popup_row += 1
            elif c == curses.KEY_SR or c == curses.KEY_UP or ch == "k":
                corner_popup_row -= 1
            elif c == curses.KEY_NPAGE or ch == ";":
                corner_popup_row += 20
            elif c == curses.KEY_PPAGE or ch == "'":
                corner_popup_row -= 20
            elif c == curses.KEY_RIGHT or ch == "l":
                corner_popup_col += 8
            elif c == curses.KEY_LEFT or ch == "h":
                corner_popup_col -= 8
            # Terminal resize signal
            # elif c == curses.KEY_RESIZE:
            elif c == ord("y") or c == ord("Y") - 32 or c == ord(" "):
                c = ord("y")
                break
            elif c == ord("n") or c == ord("N") - 32:
                c = ord("n")
                break
            elif c == ord("q") or c == ord("Q") - 32:
                c = ord("q")
                break
        return c

    def confirm_and_write_to_file(self):
        logger.debug("write to a file: imediff.out")
        c = self.display_popup_win(
            "Are you sure to write result to a file: imediff.out? (Y/n)"
        )
        if c == ord("y"):
            try:
                with open(
                    "imediff.out", mode="w", buffering=io.DEFAULT_BUFFER_SIZE
                ) as fp:
                    fp.write("\n".join(self.virt_lines))
                ret = 0
            except OSError as err:
                logger.error(
                    "Error {} in creating an output file: {}".format(err, "imediff.out")
                )
                ret = 1
        else:
            ret = -1  # skip
        return ret


if __name__ == "__main__":
    logging.basicConfig(
        format="%(levelname)s: %(filename)s: %(funcName)s: %(message)s",
        filename="example.log",
        level=logging.DEBUG,
    )
    logger.debug("__main__ start")
    tui = TUI(text_example)
    tui.main()
    # finish
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=78 sw=4 sts=4 ts=4 et ai si ft=python fileencoding=utf-8 :

"""
Module lines2lib -- matching 2 lines library

Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

"""
from difflib import SequenceMatcher

import re
import sys
import logging

logger = logging.getLogger(__name__)


class LineMatcher:
    """
    Linematcher

    A public class to help manage 2 lists of similar lines by finding fuzzy
    matching of filtered lines. This fuzzy matching of filtered lines is
    controlled by the line_rule value.

    line_rule   pattern    -- filtering behavior
         0      r""        -- drop none between text, but strip
         1      r"\\s+"     -- drop all whitespaces
         2      r"[\\s\\"']" -- drop all whitespaces and quotes (default)
         3      r"\\W+"     -- drop all non-alphanumerics

         10     r""        -- drop none between text, but strip and lowercase
         11     r"\\s+"     -- drop all whitespaces and lowercase
         12     r"[\\s\\"']" -- drop all whitespaces and quotes and lowercase
         13     r"\\W+"     -- drop all non-alphanumerics and lowercase

    Here, actual code should use a single backslash for every 2 of them.

    The fuzzy matching is performed on partial match of head and tail portion
    of lines provided by _LineMatcher class.  This partial match is controlled
    by 3 parametrs:

    parameters
    * line_max     -- initial length to compare (upper limit, defalt=128)
    * line_min     -- final   length to compare (lower limit, default=1)
    * line_factor  -- length shortening factor (default=8 for x0.8)

    The resulting tag from Linematcher class object are:

    * 'E' ----------------------- for a[j1:j2] == b[i1:i2] -- exact match
    * 'N' ----------------------- for a[j1:j2] != b[i1:i2] -- no match
    * 'F' ----------------------- for a[j1:j2] != b[i1:i2] -- fuzzy match

    Example:
    >>> a = [   "line 1 abcde\\n",
    ...         "line 2 qazws\\n",
    ...         "line 3 edcrf\\n",
    ...         "line 4 tgbyh\\n",
    ...         "line 5 yhnuj\\n",
    ...         "line 6 ujmik\\n",
    ...         "line 7 olpvb\\n",
    ...         "line 8 abcde\\n",
    ...         "line 9 olpvb\\n",
    ...         "line 0 abcde\\n",
    ...         "line 0 abcde\\n",
    ...         "line A olpvb\\n",
    ...         "line B abcde\\n",
    ...         "line Z abcde\\n",
    ...     ]
    >>> b = [   "line 1 abcde\\n",
    ...         " line 2 qazws XXX\\n",
    ...         "  line 3 e dcrf\\n",
    ...         "  line 4 tgbyh\\n",
    ...         "   line 5  yhnuj\\n",
    ...         "   line 6  uj mik\\n",
    ...         "line 7 olpvb\\n",
    ...         "   line 8 abc\\n",
    ...         "   line X XXX\\n",
    ...         "   line X XXX\\n",
    ...         "ine  9 olp vb\\n",
    ...         " l ine  0 abc\\n",
    ...         " line A  olpv b\\n",
    ...         "B  abc\\n",
    ...         "l i n e Z 'a b c d' \\\"e\\n",
    ...     ]
    >>> lines = LineMatcher(a, b)
    >>> lines._dump_opcodes()
    match: 0 -> 0, tag = E
        a: line 1 abcde
        b: line 1 abcde
    match: 1 -> 1, tag = F
        a: line 2 qazws
        b:  line 2 qazws XXX
    match: 2 -> 2, tag = F
        a: line 3 edcrf
        b:   line 3 e dcrf
    match: 3 -> 3, tag = F
        a: line 4 tgbyh
        b:   line 4 tgbyh
    match: 4 -> 4, tag = F
        a: line 5 yhnuj
        b:    line 5  yhnuj
    match: 5 -> 5, tag = F
        a: line 6 ujmik
        b:    line 6  uj mik
    match: 6 -> 6, tag = E
        a: line 7 olpvb
        b: line 7 olpvb
    match: 7 -> 7, tag = F
        a: line 8 abcde
        b:    line 8 abc
    UNmatch: 8:8 -> 8:10, tag = N
    match: 8 -> 10, tag = F
        a: line 9 olpvb
        b: ine  9 olp vb
    match: 9 -> 11, tag = F
        a: line 0 abcde
        b:  l ine  0 abc
    UNmatch: 10:11 -> 12:12, tag = N
    match: 11 -> 12, tag = F
        a: line A olpvb
        b:  line A  olpv b
    match: 12 -> 13, tag = F
        a: line B abcde
        b: B  abc
    match: 13 -> 14, tag = F
        a: line Z abcde
        b: l i n e Z 'a b c d' "e
    """

    def __init__(
        self,
        a=[],
        b=[],
        line_rule=2,
        line_max=128,  # initial length to compare (upper limit)
        line_min=1,  # final   length to compare (lower limit)
        line_factor=8,  # length shortening factor
        # 8 for 80% of length_before every 2 steps
    ):
        """
        Construct a LineMatcher object using whitespace filtered object and _LineMatcher internal object
        class which find matches including partial matches

        """

        # initialize
        self.a = a
        self.b = b
        if not (line_rule >= 0 and line_rule < 20):
            logger.error(
                "E: line_rule should be between 0 and 19 but {}".format(line_rule)
            )
            sys.exit(2)
        # line_rule:
        # 0      r""        -- drop none between text, but strip
        # 1      r"\s+"     -- drop all whitespaces
        # 2      r"[\s\"']" -- drop all whitespaces and quotes (default)
        # 3      r"\W+"     -- drop all non-alphanumerics
        # 10     r""        -- drop none between text, but strip and lowercase
        # 11     r"\s+"     -- drop all whitespaces and lowercase
        # 12     r"[\s\"']" -- drop all whitespaces and quotes and lowercase
        # 13     r"\W+"     -- drop all non-alphanumerics and lowercase
        if (line_rule % 10) == 0:
            re_preform = re.compile(r"")
        elif (line_rule % 10) == 1:
            re_preform = re.compile(r"\s+")
        elif (line_rule % 10) == 2:
            re_preform = re.compile(r"[\s\"']+")
        elif (line_rule % 10) == 3:
            re_preform = re.compile(r"\W+")
        else:
            re_preform = re.compile(r"")
        self.a_int = []
        for ax in a:
            if line_rule < 10:
                filtered_ax = re_preform.sub("", ax).strip()
            else:
                filtered_ax = re_preform.sub("", ax).strip().lower()
            self.a_int.append(filtered_ax)
        self.b_int = []
        for bx in b:
            if line_rule < 10:
                filtered_bx = re_preform.sub("", bx).strip()
            else:
                filtered_bx = re_preform.sub("", bx).strip().lower()
            self.b_int.append(filtered_bx)
        self.int = _LineMatcher(
            self.a_int,
            self.b_int,
            0,
            len(self.a_int),
            0,
            len(self.b_int),
            line_max=line_max,
            line_min=line_min,
            line_factor=line_factor,
        )

    def get_opcodes(self):
        match = []
        for tag, i1, i2, j1, j2 in self.int.get_opcodes():
            # this is match for self.int only
            if tag == "E":
                if self.a[i1] == self.b[j1]:
                    # real exact match
                    tag = "E"
                else:
                    # match after filter is fuzzy match
                    tag = "F"
            match.append((tag, i1, i2, j1, j2))
        return match

    def _dump_opcodes(self):
        """
        private function to dump internal data state of class object for
        LineMatcher class
        """
        for tag, i1, i2, j1, j2 in self.get_opcodes():
            if (i1 + 1) == i2 and (j1 + 1) == j2:
                print("match: {} -> {}, tag = {}".format(i1, j1, tag))
                print("    a: {}".format(self.a[i1]).rstrip())
                print("    b: {}".format(self.b[j1]).rstrip())
            else:
                print("UNmatch: {}:{} -> {}:{}, tag = {}".format(i1, i2, j1, j2, tag))


class _LineMatcher:
    """
    _LineMatcher

    A private class to help manage 2 lists of similar lines by finding
    matching lines including partial line matches.

    Example:
    >>> a = [   "line1abcde",
    ...         "line2qazws",
    ...         "line3edcrf",
    ...         "line4tgbyh",
    ...         "line5yhnuj",
    ...         "line6ujmik",
    ...         "line7olpvb",
    ...         "line8abcde",
    ...         "line9olpvb",
    ...         "line0abcde",
    ...         "line0abcde",
    ...         "lineAolpvb",
    ...         "lineBabcde",
    ...         "lineZabcde",
    ...     ]
    >>> b = [   "line1abcde",
    ...         "line2qazwsx",
    ...         "line3edcrf",
    ...         "line4tgbyhx",
    ...         "line5yhnuj",
    ...         "line6ujmikz",
    ...         "line7olpvb",
    ...         "line8abcdez",
    ...         "line9olpvb",
    ...         "line0abcde",
    ...         "line0abcdexx",
    ...         "lineAolpvb",
    ...         "lineBabcde",
    ...         "lineZabcdex",
    ...     ]
    >>> lines = _LineMatcher(a, b, 0, len(a), 0, len(b))
    >>> lines._dump_opcodes()
    match: 0 -> 0, tag = E
        a: line1abcde
        b: line1abcde
    match: 1 -> 1, tag = F
        a: line2qazws
        b: line2qazwsx
    match: 2 -> 2, tag = E
        a: line3edcrf
        b: line3edcrf
    match: 3 -> 3, tag = F
        a: line4tgbyh
        b: line4tgbyhx
    match: 4 -> 4, tag = E
        a: line5yhnuj
        b: line5yhnuj
    match: 5 -> 5, tag = F
        a: line6ujmik
        b: line6ujmikz
    match: 6 -> 6, tag = E
        a: line7olpvb
        b: line7olpvb
    match: 7 -> 7, tag = F
        a: line8abcde
        b: line8abcdez
    match: 8 -> 8, tag = E
        a: line9olpvb
        b: line9olpvb
    match: 9 -> 9, tag = E
        a: line0abcde
        b: line0abcde
    match: 10 -> 10, tag = F
        a: line0abcde
        b: line0abcdexx
    match: 11 -> 11, tag = E
        a: lineAolpvb
        b: lineAolpvb
    match: 12 -> 12, tag = E
        a: lineBabcde
        b: lineBabcde
    match: 13 -> 13, tag = F
        a: lineZabcde
        b: lineZabcdex
    """

    def __init__(
        self,
        a=[],
        b=[],
        is1=0,
        is2=0,
        js1=0,
        js2=0,
        depth=0,  # used to track recursive call to _LineMatcher-class
        line_max=128,  # initial length to compare (upper limit)
        line_min=1,  # final   length to compare (lower limit)
        line_factor=8,  # length shortening factor
        # 8 for 80% of length_before every 2 steps
    ):
        """
        Construct a _LineMatcher

        """

        # initialize
        self.a = a
        self.b = b
        self.is1 = is1
        self.is2 = is2
        self.js1 = js1
        self.js2 = js2
        self.depth = depth
        self.line_min = line_min
        self.line_factor = line_factor
        maxlen = 0
        for i in range(is1, is2):
            len_a = len(a[i])
            if len_a > maxlen:
                maxlen = len_a
        for j in range(js1, js2):
            len_b = len(b[j])
            if len_b > maxlen:
                maxlen = len_b
        self.line_max = min(line_max, maxlen // 2)

    def get_opcodes(self):
        if self.depth == 0:  # depth = 0
            side = 0
            logger.debug(
                "{}_LineMatcher ===  a[{}:{}]/b[{}:{}]  ===  line_full[:]  ===".format(
                    "    " * self.depth, self.is1, self.is2, self.js1, self.js2
                ),
            )
        elif self.depth > 0:  # depth > 0
            if self.depth % 2 == 1:  # depth = 1, 3, 5, ...
                side = +1
                logger.debug(
                    "{}_LineMatcher ===  a[{}:{}]/b[{}:{}]  ===  line_head[:{:02d}] ===".format(
                        "    " * self.depth,
                        self.is1,
                        self.is2,
                        self.js1,
                        self.js2,
                        self.line_max,
                    ),
                )
            else:  # self.depth % 2 == 0:  # depth = 2, 4, 6, ...
                side = -1
                logger.debug(
                    "{}_LineMatcher ===  a[{}:{}]/b[{}:{}]  ===  line_tail[-{:02d}:] ===".format(
                        "    " * self.depth,
                        self.is1,
                        self.is2,
                        self.js1,
                        self.js2,
                        self.line_max,
                    ),
                )
        else:
            logger.error(
                "{}_LineMatcher ===  a[{}:{}]/b[{}:{}]  ===  depth should not be negative".format(
                    "!!!!" * self.depth, self.is1, self.is2, self.js1, self.js2
                )
            )
            sys.exit(2)
        if side == 0:
            # self.is1, self.is2, self.js1, self.js2 are known to cover all
            side_id = "full"
            am = self.a
            bm = self.b
        elif side == 1:  # left side match (odd-depth)
            side_id = "head"
            am = []
            bm = []
            for i in range(self.is1, self.is2):
                am.append(self.a[i][: self.line_max])
            for j in range(self.js1, self.js2):
                bm.append(self.b[j][: self.line_max])
        else:  # side == -1, right side match (even-depth)
            side_id = "tail"
            am = []
            bm = []
            for i in range(self.is1, self.is2):
                am.append(self.a[i][-self.line_max :])
            for j in range(self.js1, self.js2):
                bm.append(self.b[j][-self.line_max :])
        for i in range(self.is1, self.is2):
            logger.debug(
                "{}_LineMatcher_filter a[{}] -> {}:am[{}]='{}'".format(
                    "    " * self.depth, i, side_id, i - self.is1, am[i - self.is1]
                ),
            )
        for j in range(self.js1, self.js2):
            logger.debug(
                "{}_LineMatcher_filter b[{}] -> {}:bm[{}]='{}'".format(
                    "    " * self.depth, j, side_id, j - self.js1, bm[j - self.js1]
                ),
            )
        seq = SequenceMatcher(None, am, bm)
        match = []
        for tag, i1, i2, j1, j2 in seq.get_opcodes():
            logger.debug(
                "{}<< SequenceMatcher_tag={}  ===  a[{}:{}]/b[{}:{}]".format(
                    "    " * self.depth,
                    tag,
                    i1 + self.is1,
                    i2 + self.is1,
                    j1 + self.js1,
                    j2 + self.js1,
                ),
            )
            ip1 = self.is1 + i1
            ip2 = self.is1 + i2
            jp1 = self.js1 + j1
            jp2 = self.js1 + j2
            if tag == "equal":
                # multi line section and equal for filtered lines
                for i in range(i1, i2):
                    ip = self.is1 + i
                    jp = self.js1 + (i - i1) + j1
                    if side == 0:
                        # full match on filtered line
                        match.append(("E", ip, ip + 1, jp, jp + 1))
                        logger.debug(
                            "{}>> _LineMatcher_tag=E  ===  a[{}:{}]/b[{}:{}]  === full match".format(
                                "    " * self.depth, ip, ip + 1, jp, jp + 1
                            ),
                        )
                    else:
                        # partial match only (F for fuzzy)
                        match.append(("F", ip, ip + 1, jp, jp + 1))
                        logger.debug(
                            "{}>> _LineMatcher_tag=F  ===  a[{}:{}]/b[{}:{}]  === fuzzy match".format(
                                "    " * self.depth, ip, ip + 1, jp, jp + 1
                            ),
                        )
            elif i1 == i2 or j1 == j2:
                # delete "a" or delete "b" for filtered lines
                match.append(("N", ip1, ip2, jp1, jp2))
                logger.debug(
                    "{}>> _LineMatcher_tag=N  ===  a[{}:{}]/b[{}:{}]  === delete one side".format(
                        "    " * self.depth, ip1, ip2, jp1, jp2
                    ),
                )
            elif (i1 + 1) == i2 and (j1 + 1) == j2:
                # single line match -> assume fuzzy match without checking
                # even though this is not equal
                match.append(("F", ip1, ip2, jp1, jp2))
                logger.debug(
                    "{}>> _LineMatcher_tag=N  ===  a[{}:{}]/b[{}:{}]  === fuzzy non-match (single line)".format(
                        "    " * self.depth, ip1, ip2, jp1, jp2
                    ),
                )
            else:  # fuzzy match was not resolved
                # dig deeper for multi-line changes to find fuzzy matches
                if side == 0:  # full
                    # full -> left side
                    logger.debug(
                        "{}>> _LineMatcher_tag=?  ===  a[{}:{}]/b[{}:{}]  === dig deeper depth={} from full".format(
                            "    " * self.depth, ip1, ip2, jp1, jp2, self.depth + 1
                        ),
                    )
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=ip1,
                            is2=ip2,
                            js1=jp1,
                            js2=jp2,
                            line_max=self.line_max,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                elif side == +1:  # head side
                    # left side -> right side
                    logger.debug(
                        "{}>> _LineMatcher_tag=?  ===  a[{}:{}]/b[{}:{}]  === dig deeper depth={} from full/tail".format(
                            "    " * self.depth, ip1, ip2, jp1, jp2, self.depth + 1
                        ),
                    )
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=ip1,
                            is2=ip2,
                            js1=jp1,
                            js2=jp2,
                            line_max=self.line_max,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                elif self.line_max > self.line_min:  # tail side: side == -1
                    # right side -> left side (shorter)
                    logger.debug(
                        "{}>> _LineMatcher_tag=?  ===  a[{}:{}]/b[{}:{}]  === dig deeper depth={} from tail with shorter line_max ={}".format(
                            "    " * self.depth,
                            ip1,
                            ip2,
                            jp1,
                            jp2,
                            self.depth + 1,
                            self.line_max * self.line_factor // 10,
                        ),
                    )
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=ip1,
                            is2=ip2,
                            js1=jp1,
                            js2=jp2,
                            line_max=self.line_max * self.line_factor // 10,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                else:
                    # no more shorter, give up as multi-line block change
                    logger.debug(
                        "{}>> _LineMatcher_tag=N  ===  a[{}:{}]/b[{}:{}]  === no more depth with line_max {} <= line_min {}".format(
                            "    " * self.depth,
                            ip1,
                            ip2,
                            jp1,
                            jp2,
                            self.line_max,
                            self.line_min,
                        ),
                    )
                    match.append(
                        (
                            "N",
                            ip1,
                            ip2,
                            jp1,
                            jp2,
                        )
                    )
        return match

    def _dump_opcodes(self):
        """
        private function to dump internal data state of class object for
        _LineMatcher class
        """
        for tag, i1, i2, j1, j2 in self.get_opcodes():
            if (i1 + 1) == i2 and (j1 + 1) == j2:
                print("match: {} -> {}, tag = {}".format(i1, j1, tag))
                print("    a: {}".format(self.a[i1]))
                print("    b: {}".format(self.b[j1]))
            else:
                print("UNmatch: {}:{} -> {}:{}, tag = {}".format(i1, i2, j1, j2, tag))


if __name__ == "__main__":
    import doctest

    flags = doctest.REPORT_NDIFF | doctest.FAIL_FAST
    fail, total = doctest.testmod(optionflags=flags)
    print("{} failures out of {} tests -- ".format(fail, total), end="")
    if fail == 0:
        sys.exit(0)
    else:
        sys.exit(1)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai:
"""
Test many aspect of imediff package via unittest.

  pybuild(1)
  https://docs.python.org/3/library/unittest.html#test-discovery
  In order to be compatible with test discovery, all of the test files must be
  modules or packages importable from the top-level directory of the project

  https://stackoverflow.com/questions/3295386/python-unittest-and-discovery
  The unittest filename needs to be test_*.py

To test the in-source-tree module, invoke this script in this directory as:

 $ export PYTHONPATH=$(pwd)/../src
 $ python3 test_unittest_all.py -v

To test the system installed module, invoke this script directly in this
directory as:

 $ python3 test_unittest_all.py -v

Current diff2 uses Python standard library difflib which uses a variant of
longest contiguous matching sub-sequence algorithm by Ratcliff and Obershelp
developed in the late 1980's.  If I update this imediff program to use more
modern algorithm, this test may yield slightly different result.
"""
import unittest
import subprocess
import os
import os.path
import imediff.diff3lib

# Deb package build dh_test
#
# test/ and imediff/ are copied to ../build/ directory and cwd is set to there
# test_dir = cwd_dir + "/test"
# cwd_dir + "/imediff" -- exist
# PYTHONPATH = 'PROJECT_ROOT/debian/tmp/usr/lib/python3.11/dist-packages:PROJECT_ROOT/.pybuild/cpython3_3.11/build'
# doctest_dir = "../src/imediff/"

### TEST installed system module
# Deb autopkgtest
#
# The cwd of each test is guaranteed to be the root of the source package,
# which will have been unpacked but not built. However note that the tests must
# test the installed version of the package, as opposed to programs or any
# other file from the built tree. Tests may not modify the source tree (and may
# not have write access to it).
# test_dir = cwd_dir + "/test"
# cwd_dir + "/debian" -- exist
# cwd_dir + "/src" -- exist
# cwd_dir + "/src/imediff" -- exist
# doctest_dir = "imediff/"

# Direct invocation from test_dir
# test_dir == cwd_dir
# doctest_dir = "imediff/"

# Direct invocation from PROJECT_ROOT
# test_dir = cwd_dir + "/test"
# cwd_dir + "/debian" -- exist
# cwd_dir + "/src" -- exist
# cwd_dir + "/src/imediff" -- exist
# doctest_dir = "imediff/"

##### Build time
# This is run from copied files with PYTHONPATH set
# I: PYTHONPATH = 'PROJECT_ROOT/debian/tmp/usr/lib/python3.11/dist-packages:PROJECT_ROOT/.pybuild/cpython3_3.11/build'
##### path to test directory
####print("I: test_dir ='{}'".format(test_dir))
##### I: test_dir ='PROJECT_ROOT/.pybuild/cpython3_3.11/build/test'
####
##### path to current working directory
####print("I: cwd_dir  ='{}'".format(cwd_dir))
##### I: cwd_dir  ='PROJECT_ROOT/.pybuild/cpython3_3.11/build'
##### <<< This is where imediff module is copied and located
####
##### --- subprocess.run(["ls", "-laR"])

cwd_dir = os.getcwd()
print("I: cwd_dir     = '{}'".format(cwd_dir))
print("I: test_file   = '{}' (active)".format(__file__))
test_dir = os.path.dirname(os.path.abspath(__file__))
print("I: test_dir    = '{}' (active)".format(test_dir))
base_dir = os.path.dirname(test_dir)
print("I: base_dir    = '{}'".format(base_dir))
if os.path.isdir(base_dir + "/src"):
    print("I: test in the original source tree invoked")
    doctest_dir = base_dir + "/src/imediff"
elif os.path.isdir(base_dir + "/imediff"):
    print("I: test in the copied source tree in Debian pybuild invoked")
    doctest_dir = base_dir + "/imediff"
else:
    print("I: test in an unknown environment invoked")
    exit(1)
print("I: doctest_dir = '{}'".format(doctest_dir))
if "PYTHONPATH" in os.environ:
    print("I: PYTHONPATH  = '{}'".format(os.environ["PYTHONPATH"]))
else:
    print("I: PYTHONPATH  = <undefined>")


class TestImediff(unittest.TestCase):
    a = "a12b345c6789d"
    b = "123456789"
    c = "a1234b567c89d"

    def test_diff3lib_abc(self):
        a = "a12b345c6789d"
        b = "123456789"
        c = "a1234b567c89d"
        self.assertEqual(
            imediff.diff3lib.SequenceMatcher3(a, b, c, 0, None, True).get_opcodes(),
            [
                ("e", 0, 1, 0, 0, 0, 1),
                ("E", 1, 3, 0, 2, 1, 3),
                ("A", 3, 4, 2, 2, 3, 3),
                ("E", 4, 6, 2, 4, 3, 5),
                ("C", 6, 6, 4, 4, 5, 6),
                ("E", 6, 7, 4, 5, 6, 7),
                ("A", 7, 8, 5, 5, 7, 7),
                ("E", 8, 10, 5, 7, 7, 9),
                ("C", 10, 10, 7, 7, 9, 10),
                ("E", 10, 12, 7, 9, 10, 12),
                ("e", 12, 13, 9, 9, 12, 13),
            ],
        )
        return

    def test_lines2lib_doctest(self):
        result = subprocess.call(
            "python3 " + doctest_dir + "/lines2lib.py",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_diff3lib_doctest(self):
        result = subprocess.call(
            "python3 " + doctest_dir + "/diff3lib.py",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_diff23(self):
        result = subprocess.call(
            "cd " + test_dir + ";python3 _diff23.py >z_diff23.out", shell=True
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_diff23.out z_diff23.ref >/dev/null", shell=True
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -C none --macro=w -n file_a file_b -o z_imediff2.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff2.out z_imediff2.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2_a(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -a -C none --macro=w -n file_a file_b -o z_imediff2_a.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff2_a.out z_imediff2_a.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2_b(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -b -C none --macro=w -n file_a file_b -o z_imediff2_b.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff2_b.out z_imediff2_b.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2_d(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -d -C none --macro=w -n file_a file_b -o z_imediff2_d.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff2_d.out z_imediff2_d.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2_f(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -f -C none --macro=w -n file_a file_b -o z_imediff2_f.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff2_f.out z_imediff2_f.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -C none --macro=w -n file_a file_b file_c -o z_imediff3.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3.out z_imediff3.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_a(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -a -C none --macro=w -n file_a file_b file_c -o z_imediff3_a.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_a.out z_imediff3_a.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_b(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -b -C none --macro=w -n file_a file_b file_c -o z_imediff3_b.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_b.out z_imediff3_b.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_c(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -c -C none --macro=w -n file_a file_b file_c -o z_imediff3_c.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_c.out z_imediff3_c.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_d(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -d -C none --macro=w -n file_a file_b file_c -o z_imediff3_d.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_d.out z_imediff3_d.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_f(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -f -C none --macro=w -n file_a file_b file_c -o z_imediff3_f.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_f.out z_imediff3_f.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3_g(self):
        result = subprocess.call(
            "cd "
            + test_dir
            + ";python3 _imediff.py -l -g -C none --macro=w -n file_a file_b file_c -o z_imediff3_g.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + test_dir + ";diff z_imediff3_g.out z_imediff3_g.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def tearDown(self):
        _ = subprocess.call("cd " + test_dir + ";rm -f z_*.out", shell=True)
        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

from difflib import SequenceMatcher
from imediff.utils import read_lines, write_file
from imediff.lines2lib import LineMatcher
from imediff.diff3lib import SequenceMatcher3

import tempfile
import os
import sys
import io
import time
import logging

logger = logging.getLogger(__name__)


class TextData:  # Non-TUI data
    """
        Non curses class to handle diff data for 2 or 3 list of lines

        If args.diff_mode == 2, LineMatcher is used.
        * tag = 'E' or 'N' or 'F
          * action: '=" for tag == 'E'
          * action: 'd" for tag != 'E'

        If args.diff_mode == 3, SequenceMatcher3 is used.
        * tag = 'E' or 'A' or 'C' or 'e' or 'N'
        * tag = 'n' -- special TextData tag for clean wdiff3 merge

        But when -a, -b, -c (default_action == a/b/c) is specified.
        * tag = 'E' or 'N' ('N' includes A' or 'C' or 'e' or 'N')

    Here:
        * tag is static after initialization of TextData
        * action is dynamic for tag=N/F due to the user interaction
        * action =#AC are static since these are auto set only

        The main class data self.chunk_list and its associated data
        self.usr_chunk_list are shared between
          "args.diff_mode == 2" and
          "args.diff_mode == 3"
        cases to ease common handling of them under TUI.

            self.chunk_list[chunk_index] = (
                tag,             # diff opcode tag EeNACn / ENF
                i1,              # self.list_a (start)-index
                i2,              # self.list_a (end+1)-index
                j1,              # self.list_b (start)-index
                j2,              # self.list_b (end+1)-index
                k1,              # self.list_c (start)-index or 0 for diff_mode==2
                k2,              # self.list_c (end+1)-index or 0 for diff_mode==2
                action,          # data action abcdefgAC=# (self.default_action)
                merge_buffer,    # merge_buffer for editor ([])
            ) # 9-parameter tuple

        Internally, tag is extended to add 'n' which is 1-line clean wdiff3 merge.
        Merge result is automatically stored in merge_buffer.

            self.usr_chunk_list: list of user accessible chunk_index

            content = list of strings
    """

    ####################################################################
    # Externally exposed initializer method
    ####################################################################
    def __init__(self, list_a, list_b, list_c, args, confs):
        logger.debug("starting initialization ...")
        self.list_a = list_a
        self.list_b = list_b
        self.list_c = list_c
        self.init_args(args)
        self.init_config(confs)
        self.init_chunk_list()
        logger.debug("finished initialization")
        return

    def init_args(self, args):
        self.diff_mode = args.diff_mode
        self.file_a = args.file_a
        self.file_b = args.file_b
        self.file_c = args.file_c
        self.file_o = args.output
        self.sloppy = args.sloppy
        self.isjunk = args.isjunk
        self.line_rule = args.line_rule
        self.line_min = args.line_min
        self.line_max = args.line_max
        self.line_factor = args.line_factor
        self.edit_cmd = args.edit_cmd
        self.macro = args.macro
        self.default_action = args.default_action  # 2: abdf / 3:abcdfg

    def init_config(self, confs):
        # set from confs
        if confs["config"]["confirm_exit"] != "False":
            self.confirm_exit = True
        else:
            self.confirm_exit = False
        if confs["config"]["confirm_quit"] != "False":
            self.confirm_quit = True
        else:
            self.confirm_quit = False
        self.ls0 = confs["line_separator"]["ls0"] + " %s\n"
        self.ls1 = confs["line_separator"]["ls1"] + " %s\n"
        self.ls2 = confs["line_separator"]["ls2"] + "\n"
        self.ls3 = confs["line_separator"]["ls3"] + " %s\n"
        self.ws0 = confs["word_separator"]["ws0"]
        self.ws1 = confs["word_separator"]["ws1"]
        self.ws2 = confs["word_separator"]["ws2"]
        self.ws3 = confs["word_separator"]["ws3"]
        # set up command key translation table
        # kc converts actual input command keyname to default key bindings command keyname
        # This affects terminal input only (MACRO uses system key map only)
        self.kc = dict()  # customized key code to original key code
        self.rkc = dict()  # original key chr to customized key char
        for select_key, effective_key in confs["key"].items():
            if select_key[:7] == "select_":
                typed_key = select_key[7:]
            else:
                logger.error("E: unknown select_key: {}".format(select_key))
                sys.exit(2)
            if typed_key == "":
                typed_key = " "
            if len(typed_key) > 1:
                typed_key = typed_key.upper()
            if effective_key == "":
                effective_key = " "
            self.kc[typed_key] = effective_key
            self.rkc[effective_key] = typed_key
            if (
                len(typed_key) == 1
                and len(effective_key) == 1
                and ord(typed_key) >= ord("a")
                and ord(typed_key) <= ord("z")
                and ord(effective_key) >= ord("a")
                and ord(effective_key) <= ord("z")
            ):
                cap_typed_key = typed_key.upper()
                cap_effective_key = effective_key.upper()
                self.kc[cap_typed_key] = cap_effective_key
                self.rkc[cap_effective_key] = cap_typed_key
        # Enable debug output of key remapping
        # for key, value in self.kc.items():
        #     logger.debug("kc['{}']='{}'".format(key, value))
        # for key, value in self.rkc.items():
        #     logger.debug("** rkc['{}']='{}'".format(key, value))

    def init_chunk_list(self):
        # update self.chunk_list and self.usr_chunk_list
        if self.diff_mode == 2:
            matcher_internal = LineMatcher(self.list_a, self.list_b)
            chunk_list_internal = matcher_internal.get_opcodes()
            # Set initial action to "a" or "d"
            self.chunk_list = [
                (
                    tag,
                    i1,
                    i2,
                    j1,
                    j2,
                    0,  # dummy
                    0,  # dummy
                    "",  # dummy
                    [],
                )  # chunk_list item tuple (9 param)
                for (tag, i1, i2, j1, j2) in chunk_list_internal
            ]
            # This set to "d"
            for chunk_index in range(len(self.chunk_list)):
                self.set_action(chunk_index, self.default_action)
            for chunk_index, (
                tag,
                i1,
                i2,
                j1,
                j2,
                _,
                _,
                action,
                merge_buffer,
            ) in enumerate(self.chunk_list):
                logger.debug(
                    "chunk[{}]: tag={} === a[{}:{}], b[{}:{}] === action='{}' len(merge_buffer)={}".format(
                        chunk_index,
                        tag,
                        i1,
                        i2,
                        j1,
                        j2,
                        action,
                        len(merge_buffer),
                    )
                )
        else:  # self.diff_mode == 3
            if self.default_action in ["a", "b", "c"]:
                check_same_ac = False
            else:  # "d", "f", "g"]
                check_same_ac = True
            use_LineMatcher = 1
            matcher_internal = SequenceMatcher3(
                self.list_a,
                self.list_b,
                self.list_c,
                use_LineMatcher,  # matcher
                None,  #  isjunk
                True,  # autojunk
                2,  # line_rule
                128,  # initial length to compare (upper limit)
                1,  # final   length to compare (lower limit)
                8,  # length shortening factor
                check_same_ac,  # check a vs c for tag == 'e'
            )
            chunk_list_internal = matcher_internal.get_opcodes()
            # Set initial action to "a" or "d"
            self.chunk_list = [
                (
                    tag,
                    i1,
                    i2,
                    j1,
                    j2,
                    k1,
                    k2,
                    "",  # dummy
                    [],
                )  # chunk_list item tuple (9 param)
                for (tag, i1, i2, j1, j2, k1, k2) in chunk_list_internal
            ]
            # This set to "g"
            for chunk_index in range(len(self.chunk_list)):
                self.set_action(chunk_index, self.default_action)
            for chunk_index, (
                tag,
                i1,
                i2,
                j1,
                j2,
                k1,
                k2,
                action,
                merge_buffer,
            ) in enumerate(self.chunk_list):
                logger.debug(
                    "chunk[{}]: tag={} === a[{}:{}], b[{}:{}], c[{}:{}] === action='{}' len(merge_buffer)={}".format(
                        chunk_index,
                        tag,
                        i1,
                        i2,
                        j1,
                        j2,
                        k1,
                        k2,
                        action,
                        len(merge_buffer),
                    )
                )
        if self.diff_mode == 2:
            self.usr_chunk_list = [
                chunk_index
                for chunk_index, (
                    tag,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                ) in enumerate(self.chunk_list)
                if tag != "E"
            ]
        else:  # diff_mode == 3
            if self.default_action in ["a", "b", "c"]:
                tag_resolved = ["E"]
            else:
                tag_resolved = ["E", "e", "n", "A", "C"]
            self.usr_chunk_list = [
                chunk_index
                for chunk_index, (
                    tag,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                ) in enumerate(self.chunk_list)
                if tag not in tag_resolved
            ]
        # debug
        for usr_chunk_index, chunk_index in enumerate(self.usr_chunk_list):
            logger.debug(
                "usr_chunk_index={} --> chunk[{}]".format(
                    usr_chunk_index,
                    chunk_index,
                )
            )
        if len(self.usr_chunk_list) == 0:
            self.focused_usr_chunk_index = None
        else:
            self.focused_usr_chunk_index = 0
        # save memory
        del chunk_list_internal
        del matcher_internal

    ####################################################################
    # Externally used main method
    ####################################################################
    def main(self):  # overridden for TUI by subclassing
        """Non-interactive driven by MACRO"""
        logger.debug("start with macro = '{}'".format(self.macro))
        # data update flag
        chunk_index = 0
        while True:
            ch = self.get_macro_command()
            logger.debug("macro ='{}' >> ch='{}')".format(self.macro, ch))
            if ch in ["QUIT", "q"]:
                # No prompt for CLI
                break
            elif ch in ["w", "x"] or len(self.macro) == 0:
                write_file(self.file_o, self.get_string_from_content_for_file())
                break
            else:
                # get user accessible chunk
                # Explicitly select chunk action
                chunk_index = self.get_chunk_index_from_usr_chunk_list(
                    self.focused_usr_chunk_index
                )
                logger.debug(
                    "usr_chunk_index = {}, chunk_index = {}, ch='{}'".format(
                        self.focused_usr_chunk_index, chunk_index, ch
                    ),
                )
                if ch in ["a", "b", "c", "d", "e", "f", "g"]:
                    self.set_action(chunk_index, ch)
                elif ch in ["1", "2", "3", "4", "5", "6", "7"]:
                    self.set_action(
                        chunk_index,
                        chr(ord(ch) - ord("1") + ord("a")),
                    )
                elif ch in ["A", "B", "C", "D", "E", "F", "G"]:
                    self.set_action_all(ch.lower())
                elif ch == "m":
                    self.set_updated_merge_buffer(chunk_index)
                elif ch == "M":
                    self.set_deleted_merge_buffer(chunk_index)
                elif ch == "n":
                    self.move_focus_to_any_resolvable_chunk_next()
                elif ch == "p":
                    self.move_focus_to_any_resolvable_chunk_prev()
                elif ch == "0":
                    self.move_focus_to_any_resolvable_chunk_home()
                elif ch == "9":
                    self.move_focus_to_any_resolvable_chunk_end()
                elif ch == "N":
                    self.move_focus_to_usr_chunk_next()
                elif ch == "P":
                    self.move_focus_to_usr_chunk_prev()
                elif ch == ")":
                    self.move_focus_to_usr_chunk_home()
                elif ch == "(":
                    self.move_focus_to_usr_chunk_end()
                else:
                    pass
        logger.debug("end")
        return

    ####################################################################
    # Internally used utility methods (class data get-access)
    ####################################################################
    def get_tag(self, chunk_index):
        (tag, _, _, _, _, _, _, _, _) = self.chunk_list[chunk_index]
        return tag

    def get_action(self, chunk_index):
        (_, _, _, _, _, _, _, action, _) = self.chunk_list[chunk_index]
        return action

    def get_merge_buffer(self, chunk_index):
        (_, _, _, _, _, _, _, _, merge_buffer) = self.chunk_list[chunk_index]
        return merge_buffer

    def get_content_for_chunk(self, chunk_index):
        """Return content as string based on action"""
        content = None
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            merge_buffer,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        logger.debug(
            "chunk[{}]: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}' len(merge_buffer)={}".format(
                chunk_index, tag, i1, i2, j1, j2, k1, k2, action, len(merge_buffer)
            ),
        )
        if action == "=" or action == "#":
            content = self.list_a[i1:i2]
        elif action == "a" or action == "A":
            content = self.list_a[i1:i2]
        elif action == "b" or action == "B":
            content = self.list_b[j1:j2]
        elif action == "c" or action == "C":
            content = self.list_c[k1:k2]
        elif action == "d":
            content = self.get_merge_diff(chunk_index)
        elif action == "e" or action == "G":
            if len(merge_buffer) != 0:
                content = merge_buffer
            else:
                logger.error(
                    "chunk[{}]: Bad action='{}' with missing edited merge_buffer".format(
                        chunk_index, action
                    )
                )
                sys.exit(2)
        elif action == "f" and self.diff_mode == 2 and i2 - i1 == 1 and j2 - j1 == 1:
            content = self.get_merge_wdiff2(chunk_index)
        elif (
            action == "f"
            and self.diff_mode == 3
            and i2 - i1 == 1
            and j2 - j1 == 1
            and k2 - k1 == 1
        ):  # wdiff
            (clean_merge, content) = self.get_merge_wdiff3(chunk_index)
            if clean_merge:
                logger.error(
                    "chunk[{}]: Bad action='f' for clean_merge".format(chunk_index)
                )
                sys.exit(2)
        elif (
            action == "g"
            and self.diff_mode == 3
            and i2 - i1 == 1
            and j2 - j1 == 1
            and k2 - k1 == 1
        ):
            (clean_merge, content) = self.get_merge_wdiff3(chunk_index)
            if not clean_merge:
                logger.error(
                    "chunk[{}]: Bad action='g' for unclean_merge".format(chunk_index)
                )
                sys.exit(2)
            if len(merge_buffer) != 0:
                content = merge_buffer
            else:
                logger.error(
                    "chunk[{}]: Bad action='g' with missing pre-loaded merge_buffer".format(
                        chunk_index
                    )
                )
                sys.exit(2)
        else:
            logger.error("chunk[{}]: Bad action='{}'".format(chunk_index, action))
            sys.exit(2)
        # content is at least [] (at least empty list)
        if content is None:
            logger.error("chunk[{}]: content can't be None".format(chunk_index))
            sys.exit(2)
        return content

    def get_chunk_index_from_usr_chunk_list(self, usr_chunk_index):
        if usr_chunk_index is None:
            chunk_index = None
        elif usr_chunk_index < 0:
            logger.debug(
                "underflow usr_chunk_index={}".format(usr_chunk_index),
            )
            chunk_index = None
        elif usr_chunk_index >= len(self.usr_chunk_list):
            logger.debug(
                "overflow usr_chunk_index={}".format(usr_chunk_index),
            )
            chunk_index = None
        elif len(self.usr_chunk_list) != 0:
            chunk_index = self.usr_chunk_list[usr_chunk_index]
            logger.debug(
                "usr_chunk_index={} >> chunk_index={}".format(
                    usr_chunk_index, chunk_index
                ),
            )
        else:
            chunk_index = None
        # always return valid index (including None)
        return chunk_index

    def get_merge_count(self, actions):
        """Count actions in user accessible chunk"""
        count = 0
        for usr_chunk_index in range(len(self.usr_chunk_list)):
            chunk_index = self.usr_chunk_list[usr_chunk_index]
            if self.get_action(chunk_index) in actions:
                count += 1
        return count

    def get_unresolved_count(self):
        """Count 'd' or 'f' action in user accessible chunk"""
        return self.get_merge_count("df")

    def get_macro_command(self):  # overriding for TUI
        """Macro parsing instead of curses getch"""
        if len(self.macro) == 0:
            keyname = ""  # end of MACRO and exit
        else:
            keyname = self.macro[:1]
            self.macro = self.macro[1:]
            if keyname == "[":
                pos = self.macro.find("]")
                if pos >= 0:
                    # look for ]
                    keyname = self.macro[:pos]
                    self.macro = self.macro[1 + pos :]
        if keyname == "":  # interactive
            keyname = "QUIT"  # quit w/o saving for ^C
        logger.debug("key={}".format(keyname))
        return keyname

    def get_string_from_content_for_file(self):
        """Return output of all content"""
        output = ""
        for chunk_index in range(len(self.chunk_list)):
            content = self.get_content_for_chunk(chunk_index)
            output += "".join(content)
        return output

    ####################################################################
    # Internally used utility methods (class data merge get operation)
    ####################################################################
    def get_merge_diff(self, chunk_index):
        """Return content for diff by line"""
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            merge_buffer,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        logger.debug(
            "chunk[{}]: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}' len(merge_buffer)={}".format(
                chunk_index, tag, i1, i2, j1, j2, k1, k2, action, merge_buffer
            ),
        )
        # mark up for diff display
        content = list()
        content += [self.ls0 % self.file_a]
        content += self.list_a[i1:i2]
        if self.diff_mode == 2:
            content += [self.ls2]
            content += self.list_b[j1:j2]
            content += [self.ls3 % self.file_b]
        else:  # self.diff_mode == 3
            content += [self.ls1 % self.file_b]
            content += self.list_b[j1:j2]
            content += [self.ls2]
            content += self.list_c[k1:k2]
            content += [self.ls3 % self.file_c]
        return content

    def whitespace_is_junk(self, c):
        return c in " \t"

    def get_merge_wdiff2(self, chunk_index):
        """Return content for wdiff by line (2 files)"""
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            merge_buffer,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        if i2 - i1 != 1 or j2 - j1 != 1:
            logger.error(
                "chunk[{}]: tag={} a[{}:{}]/b[{}:{}] not for wdiff2".format(
                    chunk_index, tag, i1, i2, j1, j2
                ),
            )
            sys.exit(2)
        logger.debug(
            "chunk[{}]: tag={} === a[{}:{}]/b[{}:{}]/_[{}:{}] action='{}' len(merge_buffer)={}".format(
                chunk_index, tag, i1, i2, j1, j2, k1, k2, action, len(merge_buffer)
            ),
        )
        line_a = self.list_a[i1]
        line_b = self.list_b[j1]
        if self.isjunk:
            isjunk = None
        else:
            isjunk = self.whitespace_is_junk
        matcher_internal = SequenceMatcher(isjunk, line_a, line_b, False)
        chunk_list_internal = matcher_internal.get_opcodes()
        line_string = ""
        for tag, i1, i2, j1, j2 in chunk_list_internal:
            if tag == "equal":
                line_string += line_a[i1:i2]
            else:  # other tags (mark up with word separator)
                line_string += self.ws0
                line_string += line_a[i1:i2]
                line_string += self.ws1
                line_string += line_b[j1:j2]
                line_string += self.ws3
        del chunk_list_internal
        del matcher_internal
        return [line_string]

    def get_merge_wdiff3(self, chunk_index):
        """Return content for wdiff by line (3 files)"""
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            _,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        if i2 - i1 != 1 or j2 - j1 != 1 or k2 - k1 != 1:
            logger.error(
                "chunk[{}]: tag={} === a[{}:{}]/b[{}:{}]/c[{}:{}] not for wdiff3".format(
                    chunk_index, tag, i1, i2, j1, j2, k1, k2
                ),
            )
            sys.exit(2)
        logger.debug(
            "chunk[{}]: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}' wdiff3-merge-try".format(
                chunk_index, tag, i1, i2, j1, j2, k1, k2, action
            ),
        )
        line_a = self.list_a[i1]
        line_b = self.list_b[j1]
        line_c = self.list_c[k1]
        if self.isjunk:
            isjunk = None
        else:
            isjunk = self.whitespace_is_junk
        # wdiff uses SequenceMatcher
        use_SequenceMatcher = 0
        matcher_internal = SequenceMatcher3(
            line_a,
            line_b,
            line_c,
            use_SequenceMatcher,  # matcher
            isjunk,  #  isjunk
            True,  # autojunk
            2,  # line_rule (NOT USED)
            128,  # initial length to compare (upper limit) (NOT USED)
            1,  # final   length to compare (lower limit) (NOT USED)
            8,  # length shortening factor (NOT USED)
            True,  # check a vs. c foe tag == 'e' (NOT USED)
        )
        chunk_list_internal = matcher_internal.get_opcodes()
        # logger.debug("wdiff3: \nwchunk_list_internal >>>>> {}".format(wchunk_list_internal))
        line_string = ""
        clean_merge = True
        for tag, i1, i2, j1, j2, k1, k2 in chunk_list_internal:
            if tag == "E" or tag == "e" or tag == "A":
                line_string += line_a[i1:i2]
            elif tag == "C":
                line_string += line_c[k1:k2]
            else:  # tag == "N" (mark up with word separator)
                clean_merge = False
                line_string += self.ws0
                line_string += line_a[i1:i2]
                line_string += self.ws1
                line_string += line_b[j1:j2]
                line_string += self.ws2
                line_string += line_c[k1:k2]
                line_string += self.ws3
        del matcher_internal
        del chunk_list_internal
        logger.debug(
            "chunk[{}]: clean_merge={} === tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}'".format(
                chunk_index, clean_merge, tag, i1, i2, j1, j2, k1, k2, action
            ),
        )
        return (clean_merge, [line_string])

    ####################################################################
    # Internally used utility methods (class data set-access)
    ####################################################################
    # tag:            E e A C  n     N                       (data only)
    # action request:         (g/f)  a b c d e f g    m/M   (user input)
    # action:         = # A C  E     a b c d e f
    def set_action(self, chunk_index, action_request):
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action_old,
            merge_buffer,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        if self.diff_mode == 2:
            # for 2 file merge/pick
            #   incoming tag is E or N or F
            if tag == "E":
                action = "="
            # N F
            elif action_request == "a":
                action = "a"
            elif action_request == "b":
                action = "b"
            elif action_request == "f" and i2 - i1 == 1 and j2 - j1 == 1:
                action = "f"
            elif action_request == "f":
                action = "d"
            elif action_request == "d":
                action = "d"
            elif action_request == "e" and len(merge_buffer) != 0:
                action = "e"
            elif action_request == "e" and len(merge_buffer) == 0:
                logger.warning(
                    "chunk[{}]: diff_mode={} action_request='{}' len(merge_buffer)={} --> keep action_old='{}'".format(
                        chunk_index,
                        self.diff_mode,
                        action_request,
                        len(merge_buffer),
                        action_old,
                    ),
                )
                action = action_old
            else:
                logger.warning(
                    "chunk[{}]: diff_mode={} action_request='{}' len(merge_buffer)={} action_old='{}' (very odd) --> force action='d'".format(
                        chunk_index,
                        self.diff_mode,
                        action_request,
                        len(merge_buffer),
                        action_old,
                    ),
                )
                action = "d"
        else:  # self.diff_mode == 3
            # for 3 file merge
            #   incoming tag is E or A or C or N or e
            if tag == "E":
                action = "="
            elif tag == "e" and self.default_action not in ["a", "b", "c"]:
                action = "#"
            elif tag == "n" and self.default_action not in ["a", "b", "c"]:
                action = "G"
            elif tag == "A" and self.default_action not in ["a", "b", "c"]:
                action = "A"
            elif tag == "C" and self.default_action not in ["a", "b", "c"]:
                action = "C"
            elif action_request in ["a", "b", "c", "d"]:
                action = action_request
            elif action_request == "e" and len(merge_buffer) != 0:
                action = "e"
            elif action_request == "e" and len(merge_buffer) == 0:
                logger.warning(
                    "chunk[{}]: diff_mode={} action_request='{}' len(merge_buffer)={} --> keep action_old='{}'".format(
                        chunk_index,
                        self.diff_mode,
                        action_request,
                        len(merge_buffer),
                        action_old,
                    ),
                )
                action = action_old
            elif self.default_action in ["a", "b", "c"]:
                action = self.default_action
            elif (
                action_request == "f" and i2 - i1 == 1 and j2 - j1 == 1 and k2 - k1 == 1
            ):
                # all 1 line diff -> try wdiff
                (clean_merge, content) = self.get_merge_wdiff3(chunk_index)
                if clean_merge:
                    action = "G"  # clean merge
                    tag = "n"  # update
                    merge_buffer = content
                    if self.default_action == "d":
                        self.usr_chunk_list = [
                            chunk_index
                            for chunk_index, (
                                tag,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                            ) in enumerate(self.chunk_list)
                            if tag not in ["E", "e", "n", "A", "C"]
                        ]
                        # debug
                        if len(self.usr_chunk_list) == 0:
                            self.focused_usr_chunk_index = None
                            logger.debug(
                                "reindex(f) at chunk[{}] usr_chunk[*]".format(
                                    chunk_index
                                )
                            )
                        elif self.focused_usr_chunk_index == len(self.usr_chunk_list):
                            self.focused_usr_chunk_index = len(self.usr_chunk_list) - 1
                            logger.debug(
                                "reindex(f) at chunk[{}] usr_chunk[{} (avoid overflow)]".format(
                                    chunk_index, self.focused_usr_chunk_index
                                )
                            )
                        else:
                            logger.debug(
                                "reindex(f) at chunk[{}] usr_chunk[{}]".format(
                                    chunk_index, self.focused_usr_chunk_index
                                )
                            )
                else:
                    action = "f"  # non-clean merge
            elif action_request == "f":
                action = "d"  # non-clean merge
            elif (
                action_request == "g" and i2 - i1 == 1 and j2 - j1 == 1 and k2 - k1 == 1
            ):
                # all 1 line diff -> try wdiff
                (clean_merge, content) = self.get_merge_wdiff3(chunk_index)
                if clean_merge:
                    action = "G"  # clean merge
                    tag = "n"  # update
                    merge_buffer = content
                    if self.default_action == "d":
                        self.usr_chunk_list = [
                            chunk_index
                            for chunk_index, (
                                tag,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                                _,
                            ) in enumerate(self.chunk_list)
                            if tag not in ["E", "e", "n", "A", "C"]
                        ]
                        # debug
                        if len(self.usr_chunk_list) == 0:
                            self.focused_usr_chunk_index = None
                            logger.debug(
                                "reindex(g) at chunk[{}] usr_chunk[*]".format(
                                    chunk_index
                                )
                            )
                        elif self.focused_usr_chunk_index == len(self.usr_chunk_list):
                            self.focused_usr_chunk_index = len(self.usr_chunk_list) - 1
                            logger.debug(
                                "reindex(g) at chunk[{}] usr_chunk[{} (avoid overflow)]".format(
                                    chunk_index, self.focused_usr_chunk_index
                                )
                            )
                        else:
                            logger.debug(
                                "reindex(g) at chunk[{}] usr_chunk[{}]".format(
                                    chunk_index, self.focused_usr_chunk_index
                                )
                            )
                else:
                    action = "d"  # non-clean merge
            elif action_request == "g":
                action = "d"  # non-clean merge
            elif action_request == "d":
                action = "d"
            else:  # N
                logger.warning(
                    "chunk[{}]: diff_mode={} action_request='{}' len(merge_buffer)={} action_old='{}' (very odd) --> force action='d'".format(
                        chunk_index,
                        self.diff_mode,
                        action_request,
                        len(merge_buffer),
                        action_old,
                    ),
                )
                action = "d"
        self.chunk_list[chunk_index] = (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            merge_buffer,
        )  # chunk_list item tuple (9 param)
        return

    def set_action_all(self, action_request):
        for usr_chunk_index in range(len(self.usr_chunk_list)):
            chunk_index = self.usr_chunk_list[usr_chunk_index]
            logger.debug(
                "usr_chunk_index={} >> chunk_index={} >> action_request={}".format(
                    usr_chunk_index, chunk_index, action_request
                ),
            )
            self.set_action(chunk_index, action_request)
        return

    def set_merge_buffer(self, chunk_index, merge_buffer):
        (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            _,
        ) = self.chunk_list[
            chunk_index
        ]  # chunk_list item tuple (9 param)
        self.chunk_list[chunk_index] = (
            tag,
            i1,
            i2,
            j1,
            j2,
            k1,
            k2,
            action,
            merge_buffer,
        )  # chunk_list item tuple (9 param)
        return

    def set_updated_merge_buffer(self, chunk_index):
        content = self.get_content_for_chunk(chunk_index)
        with tempfile.NamedTemporaryFile(
            mode="w",
            buffering=io.DEFAULT_BUFFER_SIZE,
            suffix=".tmp",
            prefix="imediff.",
            dir=".",
            delete=False,
        ) as fp:
            temp_file_name = fp.name
            if len(content):
                for line in content:
                    fp.write(line)
        time.sleep(0.1)  # make the change visible
        editor_ret = os.system("%s %s" % (self.edit_cmd, temp_file_name))
        time.sleep(0.1)  # make the change visible
        # time.sleep(5.0) # debug
        if editor_ret == 0:
            merge_buffer = read_lines(temp_file_name)
        else:
            merge_buffer = []  # Ignore editor errors
        os.unlink(temp_file_name)
        self.set_merge_buffer(chunk_index, merge_buffer)
        self.set_action(chunk_index, "e")
        return

    def set_deleted_merge_buffer(self, chunk_index):
        self.set_merge_buffer(chunk_index, [])
        self.set_action(chunk_index, "d")
        return

    ####################################################################
    # Internally used utility methods (active/cursor position)
    ####################################################################
    def move_focus_to_any_resolvable_chunk_next(self):
        """Jump to the next user accessible chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index_old = self.focused_usr_chunk_index
            self.focused_usr_chunk_index = min(
                self.focused_usr_chunk_index + 1,
                len(self.usr_chunk_list) - 1,
            )
            if usr_chunk_index_old == self.focused_usr_chunk_index:
                self.report("No next user accessible chunk (@end)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_any_resolvable_chunk_prev(self):
        """Jump to the previous available unresolvable chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index_old = self.focused_usr_chunk_index
            self.focused_usr_chunk_index = max(self.focused_usr_chunk_index - 1, 0)
            if usr_chunk_index_old == self.focused_usr_chunk_index:
                self.report("No previous user accessible chunk (@home)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_any_resolvable_chunk_home(self):
        """Jump to the first available unresolvable chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index_old = self.focused_usr_chunk_index
            self.focused_usr_chunk_index = 0
            if usr_chunk_index_old == self.focused_usr_chunk_index:
                self.report("No previous user accessible chunk (@home)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_any_resolvable_chunk_end(self):
        """Jump to the last available unresolvable chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index_old = self.focused_usr_chunk_index
            self.focused_usr_chunk_index = len(self.usr_chunk_list) - 1
            if usr_chunk_index_old == self.focused_usr_chunk_index:
                self.report("No next user accessible chunk (@end)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_usr_chunk_next(self):
        """Jump to the next unresolved chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index = None
            for usr_chunk_index in range(
                self.focused_usr_chunk_index + 1, len(self.usr_chunk_list)
            ):
                chunk_index = self.usr_chunk_list[usr_chunk_index]
                if self.get_action(chunk_index) in "df":
                    break
            if usr_chunk_index is not None:
                self.focused_usr_chunk_index = usr_chunk_index
            else:
                self.report("No next unselected unresolved chunk (@end)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_usr_chunk_prev(self):
        """Jump to the previous unresolved chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index = None
            for usr_chunk_index in range(self.focused_usr_chunk_index - 1, -1, -1):
                chunk_index = self.usr_chunk_list[usr_chunk_index]
                if self.get_action(chunk_index) in "df":
                    break
            if usr_chunk_index is not None:
                self.focused_usr_chunk_index = usr_chunk_index
            else:
                self.report("No previous unselected unresolved chunk (@home)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_usr_chunk_home(self):
        """Jump to the first unresolved chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index = None
            for usr_chunk_index in range(0, self.focused_usr_chunk_index):
                chunk_index = self.usr_chunk_list[usr_chunk_index]
                if self.get_action(chunk_index) in "df":
                    break
            if usr_chunk_index is not None:
                self.focused_usr_chunk_index = usr_chunk_index
            else:
                self.report("No previous unselected unresolved chunk (@home)")
        else:
            self.report("No user accessible chunk")
        return

    def move_focus_to_usr_chunk_end(self):
        """Jump to the last unselected unresolved chunk"""
        if self.focused_usr_chunk_index is not None:
            usr_chunk_index = None
            for usr_chunk_index in range(
                len(self.usr_chunk_list) - 1,
                self.focused_usr_chunk_index,
                -1,
            ):
                chunk_index = self.usr_chunk_list[usr_chunk_index]
                if self.get_action(chunk_index) in "df":
                    break
            if usr_chunk_index is not None:
                self.focused_usr_chunk_index = usr_chunk_index
            else:
                self.report("No next unselected unresolved chunk (@end)")
        else:
            self.report("No user accessible chunk")
        return

    ####################################################################
    # Internally used utility methods (user feedback)
    ####################################################################
    def report(self, message):  # override for TUI
        print(message, file=sys.stderr)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :
fp = open("report.py", "r")
line_list = fp.readlines()
#
# list of string containing \n at the end of each item
# for line in range(len(line_list)):
for line in line_list:
    print("line='{}'".format(line))

bigline = "".join(line_list)
print("================================================")
print(bigline)
print("================================================")

FOOLINE = "a\nb\nc\n"
print(FOOLINE.split())
print("================================================")
print(FOOLINE.splitlines(keepends=True))



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import io
import sys
import logging

logger = logging.getLogger(__name__)

# Utility functions
# Console width with Zenkaku=2, Hankaku=1 (Hankaku=ASCII etc.)
# Latin-1, CJK focus simplification applied.
# (For Arabic, Hebrew etc., we need some refinement)
# def console_width(text):
#     width = 0
#     for c in text:
#         if c == "\t":  # TAB = 8
#             width += 8
#         elif unicodedata.east_asian_width(c) in "FWA":  # Zenkaku
#             width += 2
#         else:
#             width += 1
#     return width


# file read
def read_lines(filename):
    logger.debug("read_lines filename = '{}'".format(filename))
    if filename is None or filename == "":
        lines = []
    else:
        try:
            with open(filename, buffering=io.DEFAULT_BUFFER_SIZE) as fp:
                lines = fp.readlines()  # read into list with tailing \n
        except Exception as _:
            lines = []
    return lines


# file output
def write_file(filename, output):
    logger.debug("write_file filename = '{}'".format(filename))
    if filename is None or filename == "-" or filename == "":
        sys.stderr.write(output)
    else:
        try:
            with open(filename, mode="w", buffering=io.DEFAULT_BUFFER_SIZE) as fp:
                fp.write(output)
        except OSError as err:
            logger.error("Error {} in creating output file: {}".format(err, filename))
            sys.exit(2)
    return


def s_number(number):
    if number is None:
        ret = "*"
    else:
        ret = "{}".format(number)
    return ret



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import os
import sys
import io
import logging

logger = logging.getLogger(__name__)

# Update version below only when configuration API changes

config_template = """\
# imediff configuration file
# Please edit right side of assignment to customize
# ini-type syntax where # starts comment
# Don't create duplicate definition
# Key side is case insensitive.

[config]
version = 3.1       # DON'T EDIT THIS.  This is for future upgrade tracking.
confirm_exit = True # Set as "False" to save and exit without pause
confirm_quit = True # Set as "False" to quit without pause
#editor = vim       # Set this to override /usr/bin/editor and $EDITOR

# key remapping is used only for TUI user input.  MACRO need to use the
# original bindings.
[key]               # action invoked by the selected key
select_a = a        # set mode a to select 'a' buffer
select_b = b        # set mode b to select 'b' buffer
select_c = c        # set mode c to select 'c' buffer (diff3)
select_d = d        # set mode d to select diff content if possible
select_e = e        # set mode e to select editor buffer if possible
select_f = f        # set mode f to select wdiff content if possible
select_g = g        # set good merge (diff3/wdiff3) if possible
select_h = h        # move left for the display scope
select_i = i
select_j = j        # move down for the display scope
select_k = k        # move up for the display scope
select_l = l        # move right for the display scope
select_m = m        # start editor to modify content / M reset
select_n = n        # move active selection to next usr_chunk
select_o = o
select_p = p        # move active selection to previous usr_chunk
select_q = q        # quit imediff without saving the result
select_r = r
select_s = s
select_t = t        # display tutorial
select_u = u
select_v = v
select_w = w        # write result and e_x_it program
select_x = x        # write result and e_x_it program
select_y = y        # key for "Yes" answer
select_z = z
select_SPACE = SPACE
select_! = !
select_" = "
select_HASH = HASH
select_$ = $
select_PERCENT = PERCENT
select_& = &
select_' = '
select_( = (
select_) = )
select_* = *
select_+ = +
select_, = ,
select_- = -
select_. = .
select_/ = /
select_0 = 0        # move active selection to the first usr_chunk
select_1 = 1        # alias for a
select_2 = 2        # alias for b
select_3 = 3        # alias for c
select_4 = 4        # alias for d
select_5 = 5        # alias for e
select_6 = 6        # alias for f
select_7 = 7        # alias for g
select_8 = 8
select_9 = 9        # move active selection to the last usr_chunk
select_COLON = COLON
select_; = ;
select_< = >
select_EQUAL = EQUAL
select_> = >
select_? = ?
select_@ = @
select_LBRACKET = LBRACKET        # MACRO block start
select_BACKSLASH = BACKSLASH
select_RBRACKET = RBRACKET        # MACRO block end
select_^ = ^
select__ = _
select_` = `
select_{ = {
select_| = |
select_} = }
select_~ = ~
select_DEL = DEL
select_TAB = TAB
select_BTAB = BTAB
select_ENTER = ENTER
select_UP = UP
select_DOWN = DOWN
select_LEFT = LEFT
select_RIGHT = RIGHT
select_INSERT = INSERT
select_DELETE = DELETE
select_HOME = HOME
select_END = END
select_PAGEUP = PAGEUP
select_PAGEDOWN = PAGEDOWN
select_BACKSPACE = BACKSPACE
select_F1 = F1
select_F2 = F2
select_F3 = F3
select_F4 = F4
select_F5 = F5
select_F6 = F6
select_F7 = F7
select_F8 = F8
select_F9 = F9
select_F10 = F10
select_F11 = F11
select_F12 = F11
#          = ^--------- customized key setting
# ^--------- reference key setting
# Upper case ASCII-keys are mapped following lower case keys

# Color with /WHITE uses white background
[attrib]
color_merge_ab           = WHITE,NORMAL         # diff2  =
color_merge_abc          = WHITE,NORMAL         # diff3  =
color_merge_ac           = WHITE,NORMAL         # diff 3 #
color_merge_a            = GREEN,NORMAL         # diff 3 A
color_merge_c            = YELLOW,NORMAL        # diff 3 C
color_merge_wdiff        = WHITE,NORMAL         # diff 3 G wdiff
color_a                  = GREEN,BOLD           # diff23 a
color_a_focus            = GREEN,BOLD,REVERSE
#color_a_focus            = GREEN/WHITE
color_b2                 = YELLOW,BOLD          # diff2  b (wdiff2)
color_b2_focus           = YELLOW,BOLD,REVERSE
#color_b2_focus           = YELLOW/WHITE
color_b3                 = MAGENTA,BOLD         # diff 3 b (wdiff3)
color_b3_focus           = MAGENTA,BOLD,REVERSE
# color_b3_focus           = MAGENTA/WHITE
color_c                  = YELLOW,BOLD          # diff 3 c
color_c_focus            = YELLOW,BOLD,REVERSE
# color_c_focus            = YELLOW/WHITE
color_editor             = CYAN,BOLD            # editor-buffer-line
color_editor_focus       = CYAN,BOLD,REVERSE
#color_editor_focus       = CYAN/WHITE
color_diff_marker        = BLUE,NORMAL          # marker-diff-line
color_diff_marker_focus  = BLUE,NORMAL,REVERSE
color_wdiff_abc          = WHITE,NORMAL         # wdiff3
color_wdiff_abc_focus    = WHITE,BOLD,REVERSE
color_wdiff_ac           = WHITE,DIM            # wdiff3
color_wdiff_ac_focus     = WHITE,BOLD,REVERSE
color_wdiff_ab           = WHITE,NORMAL         # wdiff2
color_wdiff_ab_focus     = WHITE,BOLD,REVERSE
color_wdiff_marker       = BLUE,NORMAL
color_wdiff_marker_focus = BLUE,NORMAL,REVERSE

color_status             = WHITE,NORMAL
color_status_focus       = WHITE,BOLD,REVERSE
color_white_bold         = WHITE,BOLD
color_white              = WHITE,NORMAL
color_white_reverse      = WHITE,BOLD,REVERSE
color_warn               = RED,BOLD
color_eof                = BLUE,DIM,REVERSE
color_mono               = WHITE,NORMAL


[line_separator]    # diff output formatting strings
                    # diff2 uses       ls0,      ls2, ls3
                    # diff3 uses       ls0, ls1, ls2, ls3
                    # File name  added ls0, ls1,      ls3
ls0 = <<<<<<<
ls1 = |||||||
ls2 = =======
ls3 = >>>>>>>

[word_separator]    # wdiff output formatting strings
                    # wdiff2 uses       ws0, ws1,      ws3
                    # wdiff3 uses       ws0, ws1, ws2, ws3
ws0 = {
ws1 = |
ws2 = |
ws3 = }
# alternative1 for UTF-8 terminal
#ws0 = «🅐
#ws1 = 🅑
#ws2 = 🅒
#ws3 = »
# alternative2 for UTF-8 terminal
#ws0 = «
#ws1 = ╪
#ws2 = ╫
#ws3 = »
"""


def create_template(conf):
    config_file = os.path.expanduser(conf)
    if not os.path.exists(config_file):
        logger.debug("create configuration file: {}".format(conf))
        try:
            with open(config_file, mode="w", buffering=io.DEFAULT_BUFFER_SIZE) as ofp:
                ofp.write(config_template)
        except IOError:
            logger.error("Error in creating configuration file: {}".format(conf))
            sys.exit(2)
    else:
        logger.error("Erase {} before 'imediff -t'".format(conf))
        sys.exit(2)
    return


# Generate template file: TEMPLATE.imediff
if __name__ == "__main__":
    create_template("TEMPLATE.imediff")



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
Module diff3lib -- wrapper for diff3 (generic)

Class SequenceMatcher3:
    A flexible class for comparing 3 of sequences of any type.

Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

from difflib import SequenceMatcher
from imediff.lines2lib import LineMatcher

import sys
import logging

logger = logging.getLogger(__name__)


class SequenceMatcher3:

    """
    With matcher=0, SequenceMatcher3 is an **exact** matching class for
    comparing two or three of sequences of any type, so long as the sequence
    elements are hashable using difflib imported from the system library
    SequenceMatcher.

    * For three strings, SequenceMatcher3 with matcher=0 matches these three
      strings by looking for the exact match of character pairs.
      (wdiff2/wdiff3)

    * For three lines, SequenceMatcher3 with matcher=0 matches these three
      lines by looking for the exact match of line pairs.  The resulting
      match tends to become large multi-line portion (diff2/diff3).

    With matcher=1, SequenceMatcher3 is a **fuzzy** matching class for
    comparing two or three lines using partial match pairs of filtered lines
    ignoring indentation using LineMatcher class in lines2lib.py.

    * For three lines, SequenceMatcher3 with matcher=1 matches these three
      lines by looking for the fuzzy match of line pairs. (diff2/diff3)  The
      resulting match tends to become a single-line portion suitable for
      further processing such as fuzzy mutches by SequenceMatcher3 with
      matcher=0 to obtain wdiff2/wdiff3 for that single-line.

    Let's call these three sequences as A (yours), B (base), and C (theirs).

    Let's assume A and C are derivatives from B and try to merge C into A.

    The basic rule for the merging of three sequences is as follows:

    * Check B-A matching and B-C pair matching opcode situation.
    * Walk through B to see which sides made the change. (12 cases)
    * Check if the high end of range is available (Either "E" or deletion)
    * Pick one making change as the merged content.

    * Merged as equal: 3 cases
        BA-pair opcode        | EEEENN..    EEEEEE..    EEEENN..
        BC-pair opcode        | EEEENN..    EEEENN..    EEEEEE..
        combined opcode range | ^^^^        ^^^^        ^^^^
        combined range rule   | same        BC=short    BA=short
        merge result          | A or C      A or C      A or C
        resulting tag         | E           E           E

    * Merged to pick A: 2 cases, undecided to pick conflict 1 case
        BA-pair opcode        | NNNNNN..    NNNNEE..                NNNNEE..
        BC-pair opcode        | EEEENN..    EEEENN..                EEEEEE..
        combined opcode range | ^^^^        ^^^^                    ^^^^
        combined range rule   | BC=short    same                    BA=short
        merge result          | A           A or C                  A or C
        resulting tag         | N*          A                       A

    * Merged to pick C: 2 cases, undecided to pick conflict 1 case
        BA-pair opcode        | EEEENN..                EEEENN..    EEEEEE..
        BC-pair opcode        | NNNNNN..                NNNNEE..    NNNNEE..
        combined opcode range | ^^^^                    ^^^^        ^^^^
        combined range rule   | BA=short                same        BC=short
        merge result          | C                       A or C      A or C
        resulting tag         | N*                      C           C

    * Merged as conflict: 1 case, undecided to pick conflict 2 cases
        BA-pair opcode        |            NNNNEE..    NNNNNN..    NNNNEE..
        BC-pair opcode        |            NNNNNN..    NNNNEE..    NNNNEE..
        combined opcode range |            ^^^^        ^^^^        ^^^^
        combined range rule   |            BA=short    BC=short    same
        merge result          |            C           A           A or C
        resulting tag         |            N*          N*          N

    Please note the high end of range is unavailable if the continuing side is
    "N" and not complete deletion. (marked "N*")

    All "N" and "N*" containing ranges are conflict but are checked for A==C
    check (case "e")

    Summary of tag for returned opcode tag of SequenceMatcher3

       tag    situation for 3-file-merge
     * 'E'    a[j1:j2] == b[i1:i2] == c[k1:k2] == a[j1:j2]
     * 'A'    a[j1:j2] != b[i1:i2] == c[k1:k2] != a[j1:j2]
     * 'C'    a[j1:j2] == b[i1:i2] != c[k1:k2] != a[j1:j2]
     * 'e'    a[j1:j2] != b[i1:i2] != c[k1:k2] == a[j1:j2]
     * 'N'    a[j1:j2] != b[i1:i2] != c[k1:k2] != a[j1:j2]

    For matcher=0, this uses bare SequenceMatcher class from difflib as the
    backend tool.  This uses returned tag of SequenceMatcher which are:

    * 'equal'  ------------------ for a[j1:j2] == b[i1:i2]
    * 'delete' 'insert' 'replace' for other cases

    For matcher=1, this uses LineMatcher class from diff2lib.py instead as the
    backend tool. This uses returned tag of LineMatcher which are:

    * 'E' ----------------------- for a[j1:j2] == b[i1:i2]
    * 'N' ----------------------- for a[j1:j2] != b[i1:i2]
    * 'F' ----------------------- for a[j1:j2] != b[i1:i2] -- but similar

    Example1: comparing three strings, and considering None to be "junk" and
    matcher=0.

    >>> a = "private Thread currentThread foo bar;"
    >>> b = "private volatile Thread currentThread;"
    >>> c = "private volatile currentThread foo;tail"
    >>> s = SequenceMatcher3(a, b, c, 0)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    E a[0:6],b[0:6],c[0:6] => 'privat', 'privat', 'privat'
    A a[6:6],b[6:15],c[6:15] => '', 'e volatil', 'e volatil'
    E a[6:8],b[15:17],c[15:17] => 'e ', 'e ', 'e '
    C a[8:15],b[17:24],c[17:17] => 'Thread ', 'Thread ', ''
    E a[15:28],b[24:37],c[17:30] => 'currentThread', 'currentThread', 'currentThread'
    N a[28:36],b[37:37],c[30:34] => ' foo bar', '', ' foo'
    E a[36:37],b[37:38],c[34:35] => ';', ';', ';'
    C a[37:37],b[38:38],c[35:39] => '', '', 'tail'

    Example 2: comparing three strings, and considering blanks to be "junk" and
    matcher=0.

    >>> a = "private Thread currentThread foo bar;"
    >>> b = "private volatile Thread currentThread;"
    >>> c = "private volatile currentThread foo;tail"
    >>> s = SequenceMatcher3(a, b, c, 0, lambda x: x == " ", True)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    E a[0:8],b[0:8],c[0:8] => 'private ', 'private ', 'private '
    A a[8:8],b[8:16],c[8:16] => '', 'volatile', 'volatile'
    e a[8:8],b[16:17],c[16:16] => '', ' ', ''
    C a[8:14],b[17:23],c[16:16] => 'Thread', 'Thread', ''
    E a[14:28],b[23:37],c[16:30] => ' currentThread', ' currentThread', ' currentThread'
    N a[28:36],b[37:37],c[30:34] => ' foo bar', '', ' foo'
    E a[36:37],b[37:38],c[34:35] => ';', ';', ';'
    C a[37:37],b[38:38],c[35:39] => '', '', 'tail'

    This example 2 produces more intuitive result due to setting for "junk".

    Example 3: comparing three lists with matcher=0.

    >>> a = [ "line2", "line3", "line4", "line5X", "line6X", "line7", "line8X", "line9", "line10", "line11", "line12", "xxxxxx", "aaaline14", "line15", "line16"]
    >>> b = [ "line1", "line2", "line3", "line4X", "line5", "line6X", "line7", "line8X", "line9", "line10", "line11", "line12", "yyyyyy", "line14", "line15"]
    >>> c = [ "line1", "line2", "line3X", "line4", "line5X", "line6", "line7X", "line8", "line9", "line10", "line11", "line12", "zzzzzz", "line14bbb", "line15"]
    >>> s = SequenceMatcher3(a, b, c, 0)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    A a[0:0],b[0:1],c[0:1] => '[]', '['line1']', '['line1']'
    E a[0:1],b[1:2],c[1:2] => '['line2']', '['line2']', '['line2']'
    N a[1:7],b[2:8],c[2:8] => '['line3', 'line4', 'line5X', 'line6X', 'line7', 'line8X']', '['line3', 'line4X', 'line5', 'line6X', 'line7', 'line8X']', '['line3X', 'line4', 'line5X', 'line6', 'line7X', 'line8']'
    E a[7:11],b[8:12],c[8:12] => '['line9', 'line10', 'line11', 'line12']', '['line9', 'line10', 'line11', 'line12']', '['line9', 'line10', 'line11', 'line12']'
    N a[11:13],b[12:14],c[12:14] => '['xxxxxx', 'aaaline14']', '['yyyyyy', 'line14']', '['zzzzzz', 'line14bbb']'
    E a[13:14],b[14:15],c[14:15] => '['line15']', '['line15']', '['line15']'
    A a[14:15],b[15:15],c[15:15] => '['line16']', '[]', '[]'

    Example 4: comparing three lists (same as example 3) with matcher=1.

    >>> a = [ "line2", "line3", "line4", "line5X", "line6X", "line7", "line8X", "line9", "line10", "line11", "line12", "xxxxxx", "aaaline14", "line15", "line16"]
    >>> b = [ "line1", "line2", "line3", "line4X", "line5", "line6X", "line7", "line8X", "line9", "line10", "line11", "line12", "yyyyyy", "line14", "line15"]
    >>> c = [ "line1", "line2", "line3X", "line4", "line5X", "line6", "line7X", "line8", "line9", "line10", "line11", "line12", "zzzzzz", "line14bbb", "line15"]
    >>> s = SequenceMatcher3(a, b, c, 1)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    A a[0:0],b[0:1],c[0:1] => '[]', '['line1']', '['line1']'
    E a[0:1],b[1:2],c[1:2] => '['line2']', '['line2']', '['line2']'
    C a[1:2],b[2:3],c[2:3] => '['line3']', '['line3']', '['line3X']'
    e a[2:3],b[3:4],c[3:4] => '['line4']', '['line4X']', '['line4']'
    e a[3:4],b[4:5],c[4:5] => '['line5X']', '['line5']', '['line5X']'
    C a[4:5],b[5:6],c[5:6] => '['line6X']', '['line6X']', '['line6']'
    C a[5:6],b[6:7],c[6:7] => '['line7']', '['line7']', '['line7X']'
    C a[6:7],b[7:8],c[7:8] => '['line8X']', '['line8X']', '['line8']'
    E a[7:8],b[8:9],c[8:9] => '['line9']', '['line9']', '['line9']'
    E a[8:9],b[9:10],c[9:10] => '['line10']', '['line10']', '['line10']'
    E a[9:10],b[10:11],c[10:11] => '['line11']', '['line11']', '['line11']'
    E a[10:11],b[11:12],c[11:12] => '['line12']', '['line12']', '['line12']'
    N a[11:12],b[12:13],c[12:13] => '['xxxxxx']', '['yyyyyy']', '['zzzzzz']'
    N a[12:13],b[13:14],c[13:14] => '['aaaline14']', '['line14']', '['line14bbb']'
    E a[13:14],b[14:15],c[14:15] => '['line15']', '['line15']', '['line15']'
    A a[14:15],b[15:15],c[15:15] => '['line16']', '[]', '[]'


    Methods:

    __init__(a='', b='', c='', matcher=0, isjunk=None, autojunk=True, line_rule=2)
        Construct a SequenceMatcher3.

    set_seqs(a, b)
        Set the two sequences to be compared.

    set_seq1(a)
        Set the first sequence to be compared.

    set_seq2(b)
        Set the second sequence to be compared.

    set_seq3(c)
        Set the third sequence to be compared.

    get_opcodes()
        Return list of 7-tuples describing how to merge c into a while b being
        common older version.
    """

    def __init__(
        self,
        a="",
        b="",
        c="",
        matcher=0,
        isjunk=None,
        autojunk=True,
        line_rule=2,
        line_max=128,  # initial length to compare (upper limit)
        line_min=1,  # final   length to compare (lower limit)
        line_factor=8,  # length shortening factor
        check_same_ac=True,  # check a vs. c for tag == 'e'
        # 8 for 80% of length_before every 2 steps
    ):
        """Construct a SequenceMatcher3.

        Optional arg a is the first of three sequences to be compared.  By
        default, an empty string.  The elements of a must be hashable.  See
        also .set_seqs() and .set_seq1().

        Optional arg b is the second of three sequences to be compared.  By
        default, an empty string.  The elements of b must be hashable. See
        also .set_seqs() and .set_seq2().

        Optional arg c is the third of three sequences to be compared.  By
        default, an empty string.  The elements of b must be hashable. See
        also .set_seqs() and .set_seq3().

        Optional arg matcher should be set to 0 for hashables such as strings,
        or 1 for lists of lines.

        Optional arg isjunk is None (the default), or a one-argument
        function that takes a sequence element and returns true iff the
        element is junk.  None is equivalent to passing "lambda x: 0", i.e.
        no elements are considered to be junk.  For example, pass
            lambda x: x in " \\t"
        if you're comparing lines as sequences of characters, and don't
        want to synch up on blanks or hard tabs.

        Optional arg autojunk should be set to False to disable the
        "automatic junk heuristic" that treats popular elements as junk
        (see module documentation for more information).
        """

        # Members:
        # a
        #      first sequence (yours, merge prioritize on this)
        # b
        #      second sequence (old common, used to check who changed what)
        # c
        #      third sequence (theirs, data to be merged)
        # opcodes
        #      a list of (tag, i1, i2, j1, j2, k1, k2) tuples, where tag is
        #      one of
        #          'E'    a[i1:i2] == b[j1:j2] == c[k1:k2] == a[i1:i2]
        #          'A'    a[i1:i2] != b[j1:j2] == c[k1:k2] != a[i1:i2]
        #          'C'    a[i1:i2] == b[j1:j2] != c[k1:k2] != a[i1:i2]
        #          'e'    a[i1:i2] != b[j1:j2] != c[k1:k2] == a[i1:i2]
        #          'N'    a[i1:i2] != b[j1:j2] != c[k1:k2] != a[i1:i2]
        # isjunk
        #      a user-supplied function taking a sequence element and
        #      returning true iff the element is "junk" -- this has
        #      subtle but helpful effects on the algorithm, which I'll
        #      get around to writing up someday <0.9 wink>.
        #      DON'T USE!  Only __chain_b uses this.  Use "in self.bjunk".

        self.a = a
        self.b = b
        self.c = c
        self.matcher = matcher
        self.isjunk = isjunk
        self.autojunk = autojunk
        self.line_rule = line_rule
        self.line_max = line_max
        self.line_min = line_min
        self.line_factor = line_factor
        self.check_same_ac = check_same_ac
        self.opcodes = None

    def set_seq1(self, a):
        """Set the first sequence to be compared."""

        if a is self.a:
            return
        self.a = a
        self.opcodes = None

    def set_seq2(self, b):
        """Set the second sequence to be compared."""

        if b is self.b:
            return
        self.b = b
        self.opcodes = None

    def set_seq3(self, c):
        """Set the third sequence to be compared."""

        if c is self.c:
            return
        self.c = c
        self.opcodes = None

    def set_seqs(self, a, b, c):
        """Set the two sequences to be compared."""

        self.set_seq1(a)
        self.set_seq2(b)
        self.set_seq3(c)

    def get_opcodes(self):
        """Return list of 7-tuples describing how a, b, c matches.

        Each tuple is of the form (tag, i1, i2, j1, j2, k1, k2).  The first
        tuple has i1 == j1 == k1 == 0, and remaining tuples have i1 == the i2
        from the tuple preceding it, and likewise for j1 == the previous j2,
        and likewise for k1 == the previous k2.

        The tags are strings, with these meanings:

        'E':   a == b == c
        'e':   a != b != c == a
        'A':   c == b --> a, change in a
        'C':   a == b --> c, change in c
        'N':   b-->a, b-->c, conflicting changes

        >>> a = "qabxcdsdgp"
        >>> b = "abycdfsdkg"
        >>> c = "abycdfzcpgp"
        >>> s = SequenceMatcher3(a, b, c)
        >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
        ...    print(("%s a[%d:%d] (%s) / b[%d:%d] (%s) / c[%d:%d] (%s)" %
        ...           (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2], k1, k2, c[k1:k2])))
        ...
        A a[0:1] (q) / b[0:0] () / c[0:0] ()
        E a[1:3] (ab) / b[0:2] (ab) / c[0:2] (ab)
        A a[3:4] (x) / b[2:3] (y) / c[2:3] (y)
        E a[4:6] (cd) / b[3:5] (cd) / c[3:5] (cd)
        A a[6:6] () / b[5:6] (f) / c[5:6] (f)
        N a[6:8] (sd) / b[6:9] (sdk) / c[6:9] (zcp)
        E a[8:9] (g) / b[9:10] (g) / c[9:10] (g)
        e a[9:10] (p) / b[10:10] () / c[10:11] (p)
        """

        a = self.a
        b = self.b
        c = self.c
        matcher = self.matcher
        if matcher == 0:
            opcodes_ba = SequenceMatcher(self.isjunk, b, a).get_opcodes()
            opcodes_bc = SequenceMatcher(self.isjunk, b, c).get_opcodes()
            tag_equal = "equal"
            matcher_logic = "SequenceMatcher"
        else:  # matcher == 1
            opcodes_ba = LineMatcher(
                b, a, self.line_rule, self.line_max, self.line_min, self.line_factor
            ).get_opcodes()
            opcodes_bc = LineMatcher(
                b, c, self.line_rule, self.line_max, self.line_min, self.line_factor
            ).get_opcodes()
            tag_equal = "E"
            matcher_logic = "LineMatcher"
        # index for 3-file merge
        il = jl = kl = 0  # range lower end for b, a, c (next in next round)
        ih = jh = kh = 0  # range high end for b, a, c (next in next round)
        # 2-file diff index
        n_ba = 0  # walking index for opcodes_ba
        n_bc = 0  # walking index for opcodes_bc
        len_ba = len(opcodes_ba)
        len_bc = len(opcodes_bc)
        logger.debug(
            "  matcher_logic={} tag_equal={} len_ba={} len_bc={}".format(
                matcher_logic, tag_equal, len_ba, len_bc
            ),
        )
        answer = list()
        tag = ""
        # loop start
        while n_ba < len_ba or n_bc < len_bc:
            logger.debug(
                "  loop tag_equal='{}' / j=[{}:{}] / i=[{}:{}] / k=[{}:{}] / n_ba={}, n_bc={}".format(
                    tag_equal, jl, jh, il, ih, kl, kh, n_ba, n_bc
                ),
            )
            ############################################################
            # n_ba = walking index for opcodes_ba
            ############################################################
            # (il_ba for opcodes_ba[n_ba]) == (ih_ba of previous opcodes_ba[n_ba -1]¶)
            if n_ba < len_ba:
                (tag_ba, _, ih_ba, jl_ba, jh_ba) = opcodes_ba[n_ba]
                logger.debug(
                    "      NORMAL   (n_ba={}) <  (len_ba={})".format(n_ba, len_ba)
                )
            elif len_ba == 0:  # ... underflow
                # treat as equal of zero range list at start
                (tag_ba, _, ih_ba, jl_ba, jh_ba) = (tag_equal, 0, 0, 0, 0)
                # tag_ba = tag_equal
                # ih_ba and ih_ba are the same and == 0
                # jh_ba and jh_ba are the same and == 0
                logger.debug(
                    "      UNDERRUN (n_ba={}) == (len_ba={}) == 0".format(n_ba, len_ba)
                )
            else:  # n_ba == len_ba ... overflow
                # treat as equal of zero range list at the end
                # check the last match
                (tag_ba, _, ih_ba, jl_ba, jh_ba) = opcodes_ba[len_ba - 1]
                # tag_ba = tag_equal
                # ih_ba and ih_ba are the same and == (ih_ba of the last match)
                # jh_ba and jh_ba are the same and == (jh_ba of the last match)
                (tag_ba, _, ih_ba, jl_ba, jh_ba) = (
                    tag_equal,
                    ih_ba,
                    ih_ba,
                    jh_ba,
                    jh_ba,
                )
                logger.debug(
                    "      OVERRUN  (n_ba={}) == (len_ba={}) > 0".format(n_ba, len_ba)
                )
            ############################################################
            # n_bc = walking index for opcodes_bc
            ############################################################
            # (il_bc for opcodes_bc[n_bc]) == (ih_bc of previous opcodes_ba[n_bc -1]¶)
            if n_bc < len_bc:
                (tag_bc, _, ih_bc, kl_bc, kh_bc) = opcodes_bc[n_bc]
                logger.debug(
                    "      NORMAL   (n_bc={}) <  (len_bc={})".format(n_bc, len_bc)
                )
            elif len_bc == 0:
                (tag_bc, _, ih_bc, kl_bc, kh_bc) = (tag_equal, 0, 0, 0, 0)
                logger.debug(
                    "      UNDERRUN (n_bc={}) == (len_bc={}) == 0".format(n_bc, len_bc)
                )
            else:
                (tag_bc, _, ih_bc, kl_bc, kh_bc) = opcodes_bc[len_bc - 1]
                (tag_bc, _, ih_bc, kl_bc, kh_bc) = (
                    tag_equal,
                    ih_bc,
                    ih_bc,
                    kh_bc,
                    kh_bc,
                )
                logger.debug(
                    "      OVERRUN  (n_bc={}) == (len_bc={}) > 0".format(n_bc, len_bc)
                )
            ############################################################
            # get tag for this set of opcodes if high range value is available
            ############################################################
            if tag == "N":
                # the initial value of tag is "", so only 2nd round on comes
                # here
                pass  # All undecided comes in as tag == "N", otherwise tag == ""
            elif tag_ba == tag_equal and tag_bc == tag_equal:
                tag = "E"
            elif tag_ba != tag_equal and tag_bc == tag_equal:
                tag = "A"
            elif tag_ba == tag_equal and tag_bc != tag_equal:
                tag = "C"
            elif tag_ba != tag_equal and tag_bc != tag_equal:
                tag = "N"
            ############################################################
            # walking 3-file ih to 2-file ih_*
            if ih_ba == ih_bc:
                # synched increase
                n_ba += 1
                n_bc += 1
                ih = ih_ba  # == ih_bc
                jh = jh_ba
                kh = kh_bc
            elif ih_ba > ih_bc:
                # mini-side rules 3-file ih walk
                n_bc += 1
                ih = ih_bc
                kh = kh_bc
                if tag_ba == tag_equal:
                    jh = jh_ba - (ih_ba - ih_bc)
                elif jh_ba == jl_ba:
                    jh = jh_ba
                else:
                    jh = None  # undecided
            elif ih_ba < ih_bc:
                # mini-side rules 3-file ih walk
                n_ba += 1
                ih = ih_ba
                jh = jh_ba
                if tag_bc == tag_equal:
                    # set kh to match increment step if equal
                    kh = kh_bc - (ih_bc - ih_ba)
                elif kh_bc == kl_bc:
                    # kh must be between kh_bc and kl_bc
                    kh = kh_bc
                else:
                    kh = None  # undecided
            if (jh is None) or (kh is None):
                logger.debug(
                    "    UNDECIDED tag={}, jl={}, jh={}, il={}, ih={}, kl={}, kh={}".format(
                        tag, jl, jh, il, ih, kl, kh
                    ),
                )
                # pass to next iteration
                tag = "N"
            else:
                # all determined ranges with tag
                if tag == "N" and self.check_same_ac:
                    if a[jl:jh] == c[kl:kh]:  # exact match need to be changed
                        tag = "e"
                answer.append((tag, jl, jh, il, ih, kl, kh))
                logger.debug(
                    "    APPEND    tag={}, jl={}, jh={}, il={}, ih={}, kl={}, kh={}".format(
                        tag, jl, jh, il, ih, kl, kh
                    ),
                )
                # refresh 3-file merge section
                il = ih
                jl = jh
                kl = kh
                tag = ""
        self.opcodes = answer
        return answer


if __name__ == "__main__":
    import doctest

    flags = doctest.REPORT_NDIFF | doctest.FAIL_FAST
    fail, total = doctest.testmod(optionflags=flags)
    print("{} failures out of {} tests -- ".format(fail, total), end="")
    if fail == 0:
        sys.exit(0)
    else:
        sys.exit(1)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
Copyright (C) 2024 Osamu Aoki <osamu@debian.org>
"""

import sys
import curses
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(filename="example.log", level=logging.DEBUG)
exception = None
# See https://docs.python.org/3/library/curses.html
#     https://docs.python.org/3/howto/curses.html
# logger.debug('This message should go to the log file')
# logger.info('So should this')
# logger.warning('And this, too')
# logger.error('And non-ASCII stuff, too, like Øresund and Malmö')


def test_text(char, height, width):
    test_text = ""
    for col in range(0, width):
        test_text += "{:1}".format(col % 10)
    test_text += "\n\n\n"
    for col in range(0, width):
        test_text += "{:1}".format((col // 10) % 10)
    test_text += "\n\n\n"
    for row in range(6, (height // 3) - 6):
        test_text += (char[:1] * 32 + "  {:4}  ".format(row) + char[:1] * 40) * (
            width // 80
        ) + "\n\n\n"
    for row in range(height - (height // 3) * 3, height):
        test_text += "{:1}\n".format(char[:1] * width)
    return test_text


####################################################################
# Externally used main method and effective main method gui_loop
####################################################################
def main():  # for curses wrapper
    global exception
    logger.debug("start CURSES TEST")
    curses.wrapper(gui_main)
    if exception is not None:
        print("E: {} {}".format(type(exception).__name__, exception))
    logger.debug("end CURSES TEST")
    return


def gui_main(stdscr):  # for curses TUI (core)
    global exception
    # Clear screen
    try:
        curses.curs_set(0)  # cursor off
    except Exception as e:
        exception = e
    # https://stackoverflow.com/questions/3170406/python-curses-newwin-not-working
    stdscr.clear()
    curses.curs_set(1)  # cursor on
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)
    curses.init_pair(7, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)
    # This raises ZeroDivisionError when i == 10.
    term_row_max, term_col_max = stdscr.getmaxyx()
    textarea = stdscr.derwin(term_row_max - 1, term_col_max, 0, 0)
    textarea_row_max, textarea_col_max = textarea.getmaxyx()
    statusarea = stdscr.derwin(1, term_col_max, term_row_max - 1, 0)
    statusarea_row_max, statusarea_col_max = statusarea.getmaxyx()
    test_lines = test_text("#", term_row_max - 1, term_col_max)
    for i, test_line in enumerate(test_lines.split("\n")):
        if i < term_row_max - 1:
            try:
                textarea.addstr(i, 0, test_line[:term_col_max])
            except Exception as e:
                print("E: {}".format(e))
                pass
    statusarea.addstr(
        0,
        0,
        "STEP 0: stdscr={}:{}, textarea={}:{} statusarea={}:{}".format(
            term_row_max,
            term_col_max,
            textarea_row_max,
            textarea_col_max,
            statusarea_row_max,
            statusarea_col_max,
        ),
        curses.color_pair(6),
    )
    stdscr.refresh()
    textarea.refresh()
    statusarea.refresh()
    stdscr.getkey()
    #####################################################################################
    stdscr.clear()
    stdscr.addstr(term_row_max - 1, 0, "STEP 1: ")
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################
    box_row = 12
    box_col = 60
    margin_row = get_center_margin(textarea_row_max, box_row, 5)
    margin_col = get_center_margin(textarea_col_max, box_col, 5)
    outbox = textarea.derwin(box_row, box_col, margin_row, margin_col)
    outbox_row_max, outbox_col_max = outbox.getmaxyx()
    outbox.border()
    stdscr.addstr(term_row_max - 1, 0, "STEP 2: ", curses.color_pair(2))
    outbox.refresh()
    stdscr.getkey()
    #####################################################################################
    inbox = outbox.derwin(6, 36, 1, 2)
    inbox_row_max, inbox_col_max = inbox.getmaxyx()
    inbox.addstr(
        0, 0, "Hello, world! ################################\n" * 2, curses.A_REVERSE
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 3: ", curses.color_pair(3))
    inbox.refresh()
    stdscr.getkey()
    #####################################################################################
    # intext = textarea.newpad(1000,300)  # BAD
    intext = curses.newpad(1000, 300)
    intext.addstr(0, 0, test_lines)
    intext.refresh(0, 0, 5, 5, 30, 60)
    stdscr.addstr(
        term_row_max - 1,
        0,
        "STEP 4: stdscr={}:{}, outbox={}:{}, inbox={}:{}   END of example".format(
            term_row_max,
            term_col_max,
            outbox_row_max,
            outbox_col_max,
            inbox_row_max,
            inbox_col_max,
        ),
        curses.color_pair(4),
    )
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################


def get_center_margin(span_out, span_in, margin_min):
    span_in_real = min(span_out - 2 * margin_min, span_in)
    margin = (span_out - span_in_real) // 2
    return margin


if __name__ == "__main__":
    # run program
    main()
    # finish
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""
# utility imediff functions
from imediff import __version__, __package__
from imediff.utils import read_lines
from imediff.config import create_template
from imediff.cli import TextData
from imediff.tui import TextPad
from imediff.initialize_confs import initialize_confs
from imediff.initialize_args import initialize_args

import locale
import os
import sys
import shutil
import logging

logger = logging.getLogger(__name__)

PACKAGE = __package__
version = __version__

version_string = "".join(
    filter(
        None,
        (
            """\
{p} (version {v})
""".format(
                p=PACKAGE, v=version
            ),
            __doc__,
        ),
    )
)

opening = """\
===============================================================================
                               {p}                         (version {v})

The imediff command interactively merges 2 slightly different files with an
optional base file using the in-place alternating display of the changed
content on a single-pane full screen terminal user interface (TUI).

Start this command from the shell command prompt as:
 * "imediff" to read the tutorial,
 * "imediff -h" to get all the command line options,
 * "imediff -o output OLDER NEWER" to merge 2 files, and
 * "imediff -o output MYFILE OLDFILE YOURFILE" to merge 3 files.

In the interactive TUI, you can also type "t" to read the tutorial and type "/"
to display the list of key commands.

 * Copyright (C) 2003,2004 Jarno Elonen <elonen@iki.fi>
 * Copyright (C) 2018-2025 Osamu Aoki <osamu@debian.org>

License: GPL 2.0+
===============================================================================

Type "q" to quit this program.
""".format(
    p=PACKAGE, v=version
)


##############################################################################
def main():
    """
    Entry point for imediff command

    Exit value
        0       program exits normally after saving data
        1       program quits without saving
        2       program terminates after an internal error
    """

    # preparation and arguments
    locale.setlocale(locale.LC_ALL, "")
    args = initialize_args()
    logging.basicConfig(
        format="%(levelname)s: %(filename)s: %(funcName)s: %(message)s",
        filename=args.logfile,
        level=getattr(logging, args.loglevel.upper(), logging.WARNING),
    )
    logger.debug(
        "============================== start of main =============================="
    )
    if args.template:
        create_template(args.conf)
        sys.exit(0)

    # configuration
    confs = initialize_confs(args.conf)
    for section in confs.sections():
        for key, value in confs[section].items():
            logger.debug(
                "confs['{}'] >>> key='{}' value='{}'".format(section, key, value)
            )
    editor = "editor"
    if "EDITOR" in os.environ:
        editor = os.environ["EDITOR"]
    if "editor" in confs["config"].keys():
        editor = confs["config"]["editor"]
    args.edit_cmd = shutil.which(editor)
    if args.edit_cmd is None:
        args.edit_cmd = "/usr/bin/editor"  # safe fall back
    logger.debug("external editor {} found as {}".format(editor, args.edit_cmd))

    # normalize and process non-standard situation
    if args.version:
        print(version_string)
        sys.exit(0)

    if args.diff_mode == 0:  # argument contains only zero file
        list_a = (opening).splitlines(keepends=True)
        list_b = list_a
        list_c = []
        args.file_a = ""
        args.file_b = ""
        args.diff_mode = 2
        args.macro = "t"
        confs["config"]["confirm_quit"] = "False"
        confs["config"]["confirm_exit"] = "False"
        logger.debug("=== diff0 === Tutorial for diff2 ===")
    elif args.diff_mode == 1:  # argument contains only 1 file
        list_a = (opening).splitlines(keepends=True)
        list_b = list_a
        list_c = list_a
        args.file_a = ""
        args.file_b = ""
        args.file_c = ""
        args.diff_mode = 3
        args.macro = "t"
        confs["config"]["confirm_quit"] = "False"
        confs["config"]["confirm_exit"] = "False"
        logger.debug("=== diff1 === Tutorial for diff3 ===")
    elif args.diff_mode == 2:
        # diff2
        list_a = read_lines(args.file_a)
        list_b = read_lines(args.file_b)
        list_c = None
        logger.debug(
            "=== diff2 === default_action='{}' non_interactive={} ===".format(
                args.default_action, args.non_interactive
            ),
        )
    elif args.diff_mode == 3:
        list_a = read_lines(args.file_a)
        list_b = read_lines(args.file_b)
        list_c = read_lines(args.file_c)
        logger.debug(
            "=== diff3 === default_action='{}' non_interactive={} ===".format(
                args.default_action, args.non_interactive
            ),
        )
    else:
        logger.error("imediff normally takes 2 or 3 files")
        sys.exit(2)

    if not args.non_interactive:
        display_instance = TextPad(list_a, list_b, list_c, args, confs)
        # set textpad size
        display_instance.main()
        del display_instance
    else:  # non-interactive
        text_instance = TextData(list_a, list_b, list_c, args, confs)
        text_instance.main()
        del text_instance
    logger.debug("end of main")
    sys.exit(0)
